from cms.views import details as cms_view_details
from django.core.exceptions import PermissionDenied
from django.db.models import F, Q
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse_lazy as reverse
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _

from ..forms.subjects import (
    CourseRegistrationForm,
    EventRegistrationForm,
    OrderableRegistrationForm,
    SubjectFilterForm,
    SubjectForm,
)
from ..models.courses import Course
from ..models.events import Event
from ..models.orderables import Orderable
from ..models.subjects import (
    Subject,
    SubjectPayment,
    SubjectReceivedPayment,
    SubjectRegistration,
    SubjectReturnedPayment,
    SubjectType,
)
from ..utils import reverse_with_back
from .generic import ConfirmUpdateView, CreateView, DetailView, FilteredListView, ListView, UpdateView


class SubjectTypeMixin:
    _models = {
        SubjectType.COURSE: Course,
        SubjectType.EVENT: Event,
        SubjectType.ORDERABLE: Orderable,
    }

    def dispatch(self, request, subject_type, *args, **kwargs):
        self.subject_type = get_object_or_404(SubjectType, slug=subject_type)
        self.model = self._models[self.subject_type.subject_type]
        return super().dispatch(request, *args, **kwargs)

    def get_placeholder(self):
        return super().get_placeholder() + ":" + self.subject_type.slug

    def get_queryset(self):
        return super().get_queryset().filter(subject_type=self.subject_type)

    def get_template_names(self):
        return [
            "leprikon/{}{}.html".format(self.subject_type.slug, self.template_name_suffix),
            "leprikon/{}{}.html".format(self.subject_type.subject_type, self.template_name_suffix),
            "leprikon/subject{}.html".format(self.template_name_suffix),
        ]


class CMSSubjectTypeMixin(SubjectTypeMixin):
    def dispatch(self, request, *args, **kwargs):
        # Get current CMS Page
        cms_page = request.current_page.get_public_object()

        # Check whether it really is a Leprikon subject type application.
        # It might also be one of its sub-pages.
        if cms_page.application_urls != "LeprikonSubjectTypeApp":
            # In such case show regular CMS Page
            return cms_view_details(request, *args, **kwargs)
        return super().dispatch(request, cms_page.application_namespace, **kwargs)


class SubjectListBaseView(FilteredListView):
    form_class = SubjectFilterForm
    preview_template = "leprikon/subject_preview.html"
    paginate_by = 10

    def get_title(self):
        return _("{subject_type} in school year {school_year}").format(
            subject_type=self.subject_type.plural,
            school_year=self.request.school_year,
        )

    def get_message_empty(self):
        return _("No {subject_type} matching given search parameters found.").format(
            subject_type=self.subject_type.plural,
        )

    def get_form(self):
        return self.form_class(
            subject_type_type=self.subject_type.subject_type,
            subject_types=[self.subject_type],
            school_year=self.request.school_year,
            is_staff=self.request.user.is_staff,
            data=self.request.GET,
        )


class SubjectListView(CMSSubjectTypeMixin, SubjectListBaseView):
    pass


class SubjectListMineView(SubjectTypeMixin, SubjectListBaseView):
    def get_template_names(self):
        return [
            "leprikon/{}_list_mine.html".format(self.subject_type.slug),
            "leprikon/{}_list_mine.html".format(self.subject_type.subject_type),
            "leprikon/subject_list_mine.html",
        ] + super().get_template_names()

    def get_title(self):
        return _("My {subject_type} in school year {school_year}").format(
            subject_type=self.subject_type.plural,
            school_year=self.request.school_year,
        )

    def get_queryset(self):
        return super().get_queryset().filter(leaders=self.request.leader)


class SubjectDetailView(CMSSubjectTypeMixin, DetailView):
    def get_queryset(self):
        qs = super().get_queryset()
        if not self.request.user.is_staff:
            qs = qs.filter(public=True)
        return qs


class SubjectJournalsView(SubjectTypeMixin, DetailView):
    template_name_suffix = "_journals"

    def get_queryset(self):
        qs = super().get_queryset()
        if not self.request.user.is_staff:
            qs = qs.filter(leaders=self.request.leader)
        return qs

    def get_context_data(self, **kwargs):
        return super().get_context_data(
            add_label=_("Create journal"),
            add_url=reverse_with_back(self.request, "leprikon:journal_create", kwargs={"subject": self.object.id}),
            preview_template="leprikon/journal_preview.html",
            message_empty=_("No journals matching given search parameters found."),
            **kwargs,
        )

    def get_title(self):
        return _("Journals of {subject_type} {subject}").format(
            subject_type=self.subject_type.name_genitiv,
            subject=self.object,
        )


class SubjectRegistrationsView(SubjectTypeMixin, DetailView):
    template_name_suffix = "_registrations"

    def get_queryset(self):
        qs = super().get_queryset()
        if not self.request.user.is_staff:
            qs = qs.filter(leaders=self.request.leader)
        return qs


class SubjectUpdateView(SubjectTypeMixin, UpdateView):
    form_class = SubjectForm

    def get_title(self):
        return _("Change {subject_type} {subject}").format(
            subject_type=self.subject_type.name_akuzativ,
            subject=self.object.name,
        )

    def get_message(self):
        return _("The changes in {subject_type} {subject} have been saved.").format(
            subject_type=self.subject_type.name_genitiv,
            subject=self.object.name,
        )

    def get_queryset(self):
        qs = super().get_queryset()
        if not self.request.user.is_staff:
            qs = qs.filter(leaders=self.request.leader)
        return qs


class SubjectMixin:
    def dispatch(self, request, pk, **kwargs):
        lookup_kwargs = {"subject_type": self.subject_type, "id": int(pk)}
        if not self.request.user.is_staff:
            lookup_kwargs["public"] = True
        self.subject = get_object_or_404(Subject, **lookup_kwargs)
        if not self.subject.registration_allowed:
            return redirect(self.subject)
        self.request.school_year = self.subject.school_year
        return super().dispatch(request, **kwargs)


class SubjectRegistrationFormBaseView(CreateView):
    back_url = reverse("leprikon:registration_list")
    submit_label = _("Submit registration")
    message = _("The registration has been saved. We will inform you about its further processing.")
    template_name_suffix = "_registration_form"
    _form_classes = {
        SubjectType.COURSE: CourseRegistrationForm,
        SubjectType.EVENT: EventRegistrationForm,
        SubjectType.ORDERABLE: OrderableRegistrationForm,
    }

    def get_title(self):
        return _("Registration for {subject_type} {subject}").format(
            subject_type=self.subject_type.name_akuzativ,
            subject=self.subject.name,
        )

    def get_form_class(self):
        return self._form_classes[self.subject_type.subject_type]

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["subject"] = self.subject
        kwargs["user"] = self.request.user
        return kwargs


class SubjectRegistrationFormView(CMSSubjectTypeMixin, SubjectMixin, SubjectRegistrationFormBaseView):
    pass


class UserRegistrationMixin:
    _attrs = {
        SubjectType.COURSE: "courseregistration",
        SubjectType.EVENT: "eventregistration",
        SubjectType.ORDERABLE: "orderableregistration",
    }

    def get_object(self):
        subject_registration = super().get_object()
        if subject_registration.user != self.request.user:
            raise PermissionDenied()
        return getattr(
            subject_registration,
            self._attrs[subject_registration.subject_type_type],
        )

    def get_queryset(self):
        return SubjectRegistration.objects.annotate(
            subject_type_type=F("subject__subject_type__subject_type"),
        )

    def get_template_names(self):
        return (
            [self.template_name]
            if self.template_name
            else [
                "leprikon/{}_registration{}.html".format(
                    self.object.subject.subject_type.slug, self.template_name_suffix
                ),
                "leprikon/{}_registration{}.html".format(
                    self.object.subject.subject_type.subject_type, self.template_name_suffix
                ),
                "leprikon/subject_registration{}.html".format(self.template_name_suffix),
            ]
        )


class PDFMixin:
    event = "pdf"

    def get(self, request, *args, **kwargs):
        obj = self.get_object()
        response = HttpResponse(content_type="application/pdf")
        response["Content-Disposition"] = 'attachment; filename="{}"'.format(obj.get_pdf_filename(self.event))
        return obj.write_pdf(self.event, response)


class SubjectRegistrationPdfView(PDFMixin, UserRegistrationMixin, DetailView):
    pass


class SubjectRegistrationPaymentRequestView(PDFMixin, UserRegistrationMixin, DetailView):
    event = "payment_request"


class SubjectRegistrationCancelView(UserRegistrationMixin, ConfirmUpdateView):
    template_name_suffix = "_cancel"
    title = _("Registration cancellation request")

    def get_queryset(self):
        return super().get_queryset().filter(canceled=None)

    def get_question(self):
        return _('Are you sure You want to cancel the registration "{}"?').format(self.object)

    def get_message(self):
        return _("The cancellation request for {} has been saved.").format(self.object)

    def confirmed(self):
        if self.object.approved:
            self.object.cancelation_requested = now()
            self.object.cancelation_requested_by = self.request.user
            self.object.save()
        else:
            self.object.refuse(self.request.user)


class SubjectPaymentsListView(ListView):
    model = SubjectPayment
    template_name = "leprikon/payments.html"
    paginate_by = 20

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(Q(target_registration__user=self.request.user) | Q(source_registration__user=self.request.user))
            .order_by("-accounted")
        )

    def get_title(self):
        return _("Payments")


class SubjectReceivedPaymentPdfView(PDFMixin, DetailView):
    model = SubjectReceivedPayment

    def get_queryset(self):
        return super().get_queryset().filter(target_registration__user=self.request.user)


class SubjectReturnedPaymentPdfView(PDFMixin, DetailView):
    model = SubjectReturnedPayment

    def get_queryset(self):
        return super().get_queryset().filter(source_registration__user=self.request.user)



import base64
import hashlib
import re
import secrets
import uuid
from datetime import datetime
from itertools import chain

from django.contrib.auth import get_user_model
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from django.template.loader import get_template
from django.urls import reverse
from django.utils import translation
from django.utils.functional import cached_property
from django.utils.text import slugify
from pymongo import DESCENDING, MongoClient, UpdateMany, UpdateOne
from rocketchat_API.rocketchat import RocketChat as _RocketChat

from .conf import settings
from .models.leprikonsite import LeprikonSite
from .models.messages import Message
from .models.subjects import CHAT_GROUP_BROADCAST, Subject, SubjectRegistration

User = get_user_model()

http_re = re.compile("^http")

Leprikon = {"_id": "leprikon", "username": settings.LEPRIKON_CHAT_USERNAME}


def get_rc_url():
    return LeprikonSite.objects.get_current().url + "/_chat"


def get_rc_ws_url():
    return http_re.sub("ws", get_rc_url()) + "/websocket"


class QuerySet:
    def __init__(self, model, collection, query):
        self.model = model
        self._collection = collection
        self._query = query

    def _instance(self, doc):
        instance = self.model()
        instance.__dict__.update(doc)
        return instance

    def _get(self):
        obj = self._instance(self._collection.find_one(self._query))
        if not obj:
            raise self.model.DoesNotExist()
        return obj

    def _list(self):
        return [
            self._instance(doc)
            for doc in self._collection.find(self._query,).sort(
                "ts",
                DESCENDING,
            )
        ]

    def __getitem__(self, key):
        return self._list().__getitem__(key)

    def __len__(self):
        return self._collection.count_documents(self._query)

    def _clone(self):
        return QuerySet(self.model, self._collection, self._query.copy())

    def count(self):
        return len(self)

    def filter(self, **kwargs):
        qs = self._clone()
        qs._query.update(kwargs)
        return qs

    def get(self, **kwargs):
        return self.filter(**kwargs)._get()


def get_rc_config():
    ls = LeprikonSite.objects.get_current()
    return (
        ("Accounts_iframe_enabled", True),
        ("Accounts_iframe_url", reverse("leprikon:user_login")),
        ("Accounts_Iframe_api_method", "GET"),
        ("Accounts_Iframe_api_url", reverse("leprikon:api:rocketchat")),
        ("From_Email", settings.SERVER_EMAIL_PLAIN),
        ("Organization_Name", ls.name),
        ("Organization_Email", settings.SERVER_EMAIL_PLAIN),
        ("Organization_Type", "nonprofit"),
        ("Country", "czechRepublic"),
        ("Industry", "education"),
        ("Language", 9),  # TODO: configurable language
        ("Layout_Sidenav_Footer", ""),
        ("Livechat_enabled", True),
        ("Server_Type", "privateTeam"),
        ("Register_Server", False),
        ("SMTP_Host", settings.EMAIL_HOST),
        ("SMTP_Password", settings.EMAIL_HOST_PASSWORD),
        ("SMTP_Port", settings.EMAIL_PORT),
        ("SMTP_Username", settings.EMAIL_HOST_USER),
        ("Show_Setup_Wizard", "completed"),
        ("Site_Name", ls.name),
        ("Site_Url", get_rc_url()),
        ("Website", ls.url + "/"),
    )


def get_rc_default_home(ls=None):
    if ls is None:
        ls = LeprikonSite.objects.get_current()
    with translation.override(settings.LANGUAGE_CODE):
        return get_template("rocketchat/home.html").render(
            {
                "leprikon_site": ls,
            }
        )


rc_role_permissions = (
    (
        "user",
        (
            "create-c",
            "create-d",
            "create-p",
            "create-personal-access-tokens",
            "delete-own-message",
            "leave-c",
            "leave-p",
            "mention-all",
            "mention-here",
            "view-c-room",
            "view-d-room",
            "view-history",
            "view-p-room",
            "preview-c-room",
            "view-outside-room",
            "start-discussion",
            "start-discussion-other-user",
        ),
    ),
    (
        "guest",
        (
            "create-d",
            "create-personal-access-tokens",
            "delete-own-message",
            "view-d-room",
            "view-history",
            "view-joined-room",
            "view-p-room",
            "start-discussion",
        ),
    ),
)


def get_rc_id(obj):
    if isinstance(obj, Subject):
        return f"subject-{obj.id}"
    return f"{obj.__class__.__name__.lower()}-{obj.id}"


def get_rc_u(user):
    return {"_id": get_rc_id(user), "username": user.username}


def get_rc_leprikon_data():
    return {
        "_id": "leprikon",
        "name": settings.LEPRIKON_CHAT_NAME,
        "username": settings.LEPRIKON_CHAT_USERNAME,
        "emails": [
            {
                "address": settings.SERVER_EMAIL_PLAIN,
                "verified": True,
            }
        ],
        "active": True,
        "type": "user",
        "roles": ["admin"],
        "__insert": {
            "createdAt": datetime.now(),
        },
    }


def get_rc_user_data(user):
    return {
        "_id": get_rc_id(user),
        "createdAt": user.date_joined,
        "name": user.get_full_name(),
        "username": user.username,
        "emails": [{"address": user.email, "verified": True}],
        "active": user.is_active,
        "type": "user",
        "roles": ["admin", "livechat-agent"]
        if user.is_superuser
        else (["user", "livechat-agent"] if user.is_staff else ["guest"]),
        "lastLogin": user.last_login,
        "_updatedAt": datetime.now(),
    }


def get_value(document, key):
    parts = key.split(".", 1)
    value = document[parts[0]]
    return get_value(value, parts[1]) if len(parts) == 2 else value


def get_rc_subject_room_data(subject):
    fname = f"{subject} ({subject.id})"
    now = datetime.now()
    broadcast = (subject.chat_group_type or subject.subject_type.chat_group_type) == CHAT_GROUP_BROADCAST
    return {
        "_id": get_rc_id(subject),
        "name": slugify(fname.replace("/", "-")),
        "fname": fname,
        "t": "p",
        "ro": broadcast,
        "broadcast": broadcast,
        "sysMes": False,
        "default": False,
        "u": Leprikon,
        "_updatedAt": now,
        "customFields": {"leprikon": "subject"},
        "__insert": {
            "msgs": 0,
            "usersCount": 0,
            "encrypted": False,
            "ts": now,
        },
    }


def get_rc_subscription_data(room_data, user, alert, groupMentions, owner_ids=None):
    now = datetime.now()
    subscription_data = {
        "_id": f'{room_data["_id"]}-{get_rc_id(user)}',
        "rid": room_data["_id"],
        "name": room_data["name"],
        "fname": room_data["fname"],
        "t": room_data["t"],
        "u": get_rc_u(user),
        "customFields": {"leprikon": True},
        "_updatedAt": now,
        "__insert": {
            "alert": alert,
            "open": True,
            "ts": now,
            "ls": now,
            "unread": 0,
            "userMentions": 0,
            "groupMentions": groupMentions,
        },
    }
    if owner_ids:
        subscription_data["roles"] = ["owner"] if user.id in owner_ids else []
    return subscription_data


def get_rc_subject_subscriptions_data(room_data, subject):
    leader_users = {lu.user.id: lu.user for lu in subject.leaders.select_related("user")}
    return chain(
        get_rc_subscription_data(room_data, user, False, 0, leader_users)
        for user in set(
            chain(
                leader_users.values(),
                (r.user for r in subject.registrations.select_related("user")),
            )
        )
    )


def rc_logout(auth_token, user_id):
    return (
        _RocketChat(
            auth_token=auth_token,
            user_id=user_id,
            server_url=settings.ROCKETCHAT_API_URL,
        )
        .logout()
        .json()
    )


class RocketChat:

    # mongo related helpers
    _mongo_session = None
    _mongo_trx_level = 0

    def __enter__(self):
        if not self._mongo_trx_level:
            self._mongo_session = self._mongo_client.start_session().__enter__()
            self._mongo_trx = self._mongo_session.start_transaction().__enter__()
        self._mongo_trx_level += 1
        return self

    def __exit__(self, *exc):
        self._mongo_trx_level -= 1
        if not self._mongo_trx_level:
            self._mongo_trx.__exit__(*exc)
            self._mongo_session.__exit__(*exc)
            self._mongo_session = None

    @cached_property
    def _mongo_client(self):
        return MongoClient(settings.MONGO_URL)

    @cached_property
    def _mongo(self):
        return self._mongo_client.get_default_database()

    def _get_query(self, document):
        on_insert = document.pop("__insert", False)
        query = {"$set": document}
        if on_insert:
            query["$setOnInsert"] = on_insert
        return query

    def _sync_one(self, collection, document, keys=("_id",)):
        return (
            getattr(self._mongo, collection)
            .update_one(
                {key: get_value(document, key) for key in keys},
                self._get_query(document),
                session=self._mongo_session,
                upsert=True,
            )
            .raw_result
        )

    def _sync_many(self, collection, documents, keys=("_id",)):
        operations = [
            UpdateOne(
                {key: get_value(document, key) for key in keys},
                self._get_query(document),
                upsert=True,
            )
            for document in documents
        ]
        if operations:
            return (
                getattr(self._mongo, collection)
                .bulk_write(
                    operations,
                    session=self._mongo_session,
                )
                .bulk_api_result
            )

    # RC proxy
    @cached_property
    def rc(self):
        self.sync_leprikon_user()
        return _RocketChat(
            user_id="leprikon",
            auth_token=self.create_login_token("leprikon"),
            server_url=settings.ROCKETCHAT_API_URL,
        )

    # Leprikon RC API
    def configure(self):
        return self._mongo.rocketchat_settings.bulk_write(
            [UpdateOne({"_id": key}, {"$set": {"value": value, "blocked": True}}) for key, value in get_rc_config()]
            + [
                # One time settings
                UpdateOne(
                    {"_id": "Layout_Home_Body", "value": {"$regex": "Welcome to Rocket.Chat"}},
                    {"$set": {"value": get_rc_default_home()}},
                ),
            ],
            session=self._mongo_session,
        ).bulk_api_result

    def create_login_token(self, user_id):
        login_token = secrets.token_hex()
        h = hashlib.sha256()
        h.update(login_token.encode())
        hashed_token = base64.encodebytes(h.digest()).decode().strip()
        self._mongo.users.update_one(
            {"_id": user_id},
            {
                "$push": {
                    "services.resume.loginTokens": {
                        "when": datetime.now(),
                        "hashedToken": hashed_token,
                    }
                }
            },
            session=self._mongo_session,
        )
        return login_token

    def sync_leprikon_user(self):
        return self._sync_one("users", get_rc_leprikon_data())

    def sync_user(self, user):
        return self._sync_one("users", get_rc_user_data(user))

    def sync_users(self):
        return self._sync_many("users", map(get_rc_user_data, User.objects.iterator()))

    def sync_roles(self):
        return self._mongo.rocketchat_permissions.bulk_write(
            list(
                chain.from_iterable(
                    (
                        # add missing permissions
                        UpdateMany(
                            {
                                "_id": {"$in": permissions},
                                "roles": {"$ne": role},
                            },
                            {"$push": {"roles": role}},
                        ),
                        # remove other permissions
                        UpdateMany(
                            {
                                "_id": {"$nin": permissions},
                                "roles": role,
                            },
                            {"$pull": {"roles": role}},
                        ),
                    )
                    for role, permissions in rc_role_permissions
                )
            ),
            session=self._mongo_session,
        ).bulk_api_result

    def sync_subject(self, subject):
        room_data = get_rc_subject_room_data(subject)
        with self:
            return (
                self._sync_one("rocketchat_room", room_data),
                self._sync_many("rocketchat_subscription", get_rc_subject_subscriptions_data(room_data, subject)),
            )

    def sync_subjects(self):
        room_data_subjects = [
            (get_rc_subject_room_data(subject), subject)
            for subject in Subject.objects.select_related("subject_type").iterator()
        ]
        with self:
            return (
                self._sync_many("rocketchat_room", (rds[0] for rds in room_data_subjects)),
                self._sync_many(
                    "rocketchat_subscription",
                    chain.from_iterable(get_rc_subject_subscriptions_data(*rds) for rds in room_data_subjects),
                ),
            )

    def sync_subscription(self, registration):
        room_data = get_rc_subject_room_data(registration.subject)
        return self._sync_one(
            "rocketchat_subscription",
            get_rc_subscription_data(room_data, registration.user, False, 0),
        )

    def merge_users(self, source, target):
        source_id = get_rc_id(source)
        target_id = get_rc_id(target)
        target_u = get_rc_u(target)
        with self:
            # merge rocketchat_livechat_department_agents
            for department_agent in self._mongo.rocketchat_livechat_department_agents.find(
                {"agentId": source_id}, session=self._mongo_session
            ):
                self._mongo.rocketchat_livechat_department_agents.count_documents(
                    {
                        "agentId": target_id,
                        "departmentId": department_agent["departmentId"],
                    }
                ) or self._sync_one(
                    "rocketchat_livechat_department_agents",
                    {
                        "_id": department_agent["_id"],
                        "agentId": target_id,
                        "username": target.username,
                    },
                )
            # merge rocketchat_subscription
            for subscription in self._mongo.rocketchat_subscription.find(
                {"u._id": source_id}, session=self._mongo_session
            ):
                subscription["_id"] = f"""{subscription['rid']}-{target_u['_id']}"""
                subscription["u"] = target_u
                self._sync_one(
                    "rocketchat_subscription",
                    subscription,
                )
            # merge transfer messages and rooms
            for collection in (
                self._mongo.rocketchat_message,
                self._mongo.rocketchat_room,
            ):
                collection.update_many(
                    {"u._id": source_id},
                    {"$set": {"u._id": target_id}},
                    session=self._mongo_session,
                )

    def get_leprikon_private_rooms_queryset(self):
        return QuerySet(
            Message,
            self._mongo.rocketchat_room,
            {
                "t": "p",
                "customFields.leprikon": {"$in": ["subject", "broadcast"]},
            },
        )

    def get_broadcast_room_choices(self):
        return (
            (r["_id"], r["fname"])
            for r in self._mongo.rocketchat_room.find(
                {
                    "t": "p",
                    "broadcast": True,
                    "customFields.leprikon": "broadcast",
                },
                ("_id", "fname"),
            )
        )

    def subscribe_users_to_broadcast_room(self, rid, users):
        with self:
            room_data = self._mongo.rocketchat_room.find_one(
                {
                    "_id": rid,
                    "t": "p",
                    "broadcast": True,
                    "customFields.leprikon": "broadcast",
                },
                session=self._mongo_session,
            )
            self._sync_many(
                "rocketchat_subscription",
                (get_rc_subscription_data(room_data, user, True, 1) for user in users),
                keys=("rid", "u._id"),
            )
            self._sync_one(
                "rocketchat_room",
                {
                    "_id": rid,
                    "usersCount": self._mongo.rocketchat_subscription.count_documents({"rid": rid}),
                },
            )

    def create_broadcast_room(self, author, subject, text, recipients):
        with self:
            name = slugify(subject.replace("/", "-"))
            now = datetime.now()
            room_data = {
                "_id": uuid.uuid4().hex[:10],
                "name": name,
                "fname": subject,
                "t": "p",
                "ro": True,
                "sysMes": False,
                "default": False,
                "u": get_rc_u(author),
                "_updatedAt": now,
                "msgs": 1,
                "usersCount": 0,
                "customFields": {"leprikon": "broadcast"},
                "broadcast": True,
                "encrypted": False,
                "ts": now,
            }
            self._mongo.rocketchat_room.insert_one(room_data, session=self._mongo_session)
            self._mongo.rocketchat_message.insert_one(
                {
                    "_id": room_data["_id"],
                    "rid": room_data["_id"],
                    "msg": text,
                    "ts": room_data["ts"],
                    "u": room_data["u"],
                    "mentions": [
                        {
                            "_id": "all",
                            "username": "all",
                        }
                    ],
                    "channels": [],
                    "_updatedAt": now,
                },
                session=self._mongo_session,
            )
            self.subscribe_users_to_broadcast_room(room_data["_id"], recipients)


@receiver(post_save)
def rc_sync_object(instance, **kwargs):
    if isinstance(instance, User):
        RocketChat().sync_user(instance)
    elif isinstance(instance, Subject):
        RocketChat().sync_subject(instance)
    elif isinstance(instance, LeprikonSite):
        RocketChat().configure()
    elif isinstance(instance, SubjectRegistration):
        RocketChat().sync_subscription(instance)


@receiver(post_delete)
def rc_delete_object(instance, **kwargs):
    if isinstance(instance, User):
        RocketChat().rc.users_delete(get_rc_id(instance))
    elif isinstance(instance, Subject):
        RocketChat().rc.channels_delete(get_rc_id(instance))



from django import forms
from django.contrib import admin
from django.contrib.admin.widgets import AdminRadioSelect, FilteredSelectMultiple
from django.contrib.auth import get_user_model
from django.contrib.staticfiles.templatetags.staticfiles import static
from django.db.models import F
from django.shortcuts import render
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _

from ..forms.courses import CourseDiscountAdminForm, CourseRegistrationAdminForm
from ..models.courses import Course, CourseDiscount, CourseRegistration, CourseRegistrationHistory
from ..models.schoolyear import SchoolYear, SchoolYearDivision
from ..models.subjects import SubjectType
from ..utils import currency
from .pdf import PdfExportAdminMixin
from .subjects import SubjectBaseAdmin, SubjectDiscountBaseAdmin, SubjectRegistrationBaseAdmin


@admin.register(Course)
class CourseAdmin(SubjectBaseAdmin):
    subject_type_type = SubjectType.COURSE
    registration_model = CourseRegistration
    list_display = (
        "id",
        "code",
        "name",
        "subject_type",
        "get_groups_list",
        "get_leaders_list",
        "get_times_list",
        "place",
        "public",
        "registration_allowed_icon",
        "get_registrations_link",
        "journals_link",
        "icon",
        "note",
    )
    list_export = (
        "id",
        "school_year",
        "code",
        "name",
        "department",
        "subject_type",
        "registration_type",
        "get_groups_list",
        "get_leaders_list",
        "get_age_groups_list",
        "get_target_groups_list",
        "get_times_list",
        "place",
        "public",
        "price",
        "min_participants_count",
        "max_participants_count",
        "min_group_members_count",
        "max_group_members_count",
        "min_registrations_count",
        "max_registrations_count",
        "get_approved_registrations_count",
        "get_unapproved_registrations_count",
        "note",
    )
    actions = (
        "publish",
        "unpublish",
        "copy_to_school_year",
    )

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        if "school_year_division" in form.base_fields:
            # school year division choices
            school_year_division_choices = form.base_fields["school_year_division"].widget.widget.choices
            school_year_division_choices.queryset = SchoolYearDivision.objects.filter(
                school_year=obj.school_year if obj else request.school_year,
            )
            form.base_fields["school_year_division"].choices = school_year_division_choices
        return form

    def publish(self, request, queryset):
        Course.objects.filter(id__in=[reg["id"] for reg in queryset.values("id")]).update(public=True)
        self.message_user(request, _("Selected courses were published."))

    publish.short_description = _("Publish selected courses")

    def unpublish(self, request, queryset):
        Course.objects.filter(id__in=[reg["id"] for reg in queryset.values("id")]).update(public=False)
        self.message_user(request, _("Selected courses were unpublished."))

    unpublish.short_description = _("Unpublish selected courses")

    def copy_to_school_year(self, request, queryset):
        class SchoolYearForm(forms.Form):
            school_year = forms.ModelChoiceField(
                label=_("Target school year"),
                help_text=_("All selected courses will be copied to selected school year."),
                queryset=SchoolYear.objects.all(),
            )

        if request.POST.get("post", "no") == "yes":
            form = SchoolYearForm(request.POST)
            if form.is_valid():
                school_year = form.cleaned_data["school_year"]
                for course in queryset.all():
                    course.copy_to_school_year(school_year)
                self.message_user(request, _("Selected courses were copied to school year {}.").format(school_year))
                return
        else:
            form = SchoolYearForm()

        adminform = admin.helpers.AdminForm(
            form,
            [(None, {"fields": list(form.base_fields)})],
            {},
            None,
            model_admin=self,
        )

        return render(
            request,
            "leprikon/admin/action_form.html",
            dict(
                title=_("Select target school year"),
                opts=self.model._meta,
                adminform=adminform,
                media=self.media + adminform.media,
                action="copy_to_school_year",
                action_checkbox_name=admin.helpers.ACTION_CHECKBOX_NAME,
                select_across=request.POST["select_across"],
                selected=request.POST.getlist(admin.helpers.ACTION_CHECKBOX_NAME),
            ),
        )

    copy_to_school_year.short_description = _("Copy selected courses to another school year")

    def get_message_recipients(self, request, queryset):
        return (
            get_user_model()
            .objects.filter(
                leprikon_registrations__subject__in=queryset,
            )
            .distinct()
        )


class CourseRegistrationHistoryInlineAdmin(admin.TabularInline):
    model = CourseRegistrationHistory
    extra = 0
    readonly_fields = ("course",)

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(CourseRegistration)
class CourseRegistrationAdmin(PdfExportAdminMixin, SubjectRegistrationBaseAdmin):
    form = CourseRegistrationAdminForm
    subject_type_type = SubjectType.COURSE
    actions = ("add_discounts",)
    inlines = SubjectRegistrationBaseAdmin.inlines + (CourseRegistrationHistoryInlineAdmin,)

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .prefetch_related(
                "course_registration_periods",
            )
        )

    def add_discounts(self, request, queryset):
        ABSOLUTE = "A"
        RELATIVE = "R"

        class DiscountBaseForm(forms.Form):
            discount_type = forms.ChoiceField(
                label=_("Discount type"),
                choices=((ABSOLUTE, _("absolute amount")), (RELATIVE, _("relative amount in percents"))),
                widget=AdminRadioSelect(),
            )
            amount = forms.DecimalField(label=_("Amount or number of percents"))
            explanation = CourseDiscount._meta.get_field("explanation").formfield()

        school_year_divisions = {
            school_year_division.id: school_year_division
            for school_year_division in SchoolYearDivision.objects.filter(
                courses__registrations__in=queryset,
            ).distinct()
        }

        label = _("Periods of school year division: {school_year_division}")
        DiscountForm = type(
            "DiscountForm",
            (DiscountBaseForm,),
            {
                f"periods_{school_year_division.id}": forms.ModelMultipleChoiceField(
                    label=label.format(
                        school_year_division=school_year_division.name,
                    ),
                    queryset=school_year_division.periods.all(),
                    widget=FilteredSelectMultiple(
                        verbose_name=label.format(
                            school_year_division=school_year_division.name,
                        ),
                        is_stacked=False,
                    ),
                )
                for school_year_division in school_year_divisions.values()
            },
        )
        if request.POST.get("post", "no") == "yes":
            form = DiscountForm(request.POST)
            if form.is_valid():
                CourseDiscount.objects.bulk_create(
                    CourseDiscount(
                        registration=registration,
                        registration_period=registration_period,
                        amount=(
                            form.cleaned_data["amount"]
                            if form.cleaned_data["discount_type"] == ABSOLUTE
                            else (
                                registration.price
                                - sum(
                                    discount.amount
                                    for discount in registration.all_discounts
                                    if discount.registration_period_id == registration_period.id
                                )
                            )
                            * form.cleaned_data["amount"]
                            / 100
                        ),
                        explanation=form.cleaned_data["explanation"],
                    )
                    for registration in queryset.annotate(
                        school_year_division_id=F("subject__course__school_year_division_id")
                    )
                    for registration_period in registration.course_registration_periods.filter(
                        period__in=form.cleaned_data[f"periods_{registration.school_year_division_id}"]
                    )
                )
                self.message_user(request, _("The discounts have been created for selected registrations."))
                return
        else:
            form = DiscountForm()

        adminform = admin.helpers.AdminForm(
            form,
            [(None, {"fields": list(form.base_fields)})],
            {},
            None,
            model_admin=self,
        )

        return render(
            request,
            "leprikon/admin/action_form.html",
            dict(
                title=_("Add discounts"),
                opts=self.model._meta,
                adminform=adminform,
                media=self.media + adminform.media,
                action="add_discounts",
                action_checkbox_name=admin.helpers.ACTION_CHECKBOX_NAME,
                select_across=request.POST["select_across"],
                selected=request.POST.getlist(admin.helpers.ACTION_CHECKBOX_NAME),
            ),
        )

    add_discounts.short_description = _("Add discounts to selected registrations")

    def discounts(self, obj: CourseRegistration):
        html = []
        for status in obj.period_payment_statuses:
            query = f"?registration={obj.id}&registration_period={status.registration_period.id}"
            html.append(
                format_html(
                    '{period}: <a href="{href}"><b>{amount}</b></a>',
                    period=status.registration_period.period.name,
                    href=reverse("admin:leprikon_coursediscount_changelist") + query,
                    amount=currency(status.status.discount),
                )
                + format_html(
                    '<a class="popup-link" href="{href}" style="background-position: 0 0" title="{title}">'
                    '<img src="{icon}" alt="+"/></a>',
                    href=reverse("admin:leprikon_coursediscount_add") + query,
                    title=_("add discount"),
                    icon=static("admin/img/icon-addlink.svg"),
                )
            )
        return "<br/>".join(html)

    discounts.allow_tags = True
    discounts.short_description = _("discounts")

    def total_price(self, obj: CourseRegistration):
        html = []
        for status in obj.period_payment_statuses:
            html.append(
                format_html(
                    "{period}: <b>{amount}</b>",
                    period=status.registration_period.period.name,
                    amount=currency(status.status.receivable),
                )
            )
        return "<br/>".join(html)

    total_price.allow_tags = True
    total_price.short_description = _("total price")

    def received_payments(self, obj: CourseRegistration):
        html = []
        for period in obj.period_payment_statuses:
            html.append(
                format_html(
                    '{period}: <a style="color: {color}" href="{href}" title="{title}">' "<b>{amount}</b></a>",
                    period=period.registration_period.period.name,
                    color=period.status.color,
                    href=reverse("admin:leprikon_subjectreceivedpayment_changelist") + f"?target_registration={obj.id}",
                    title=period.status.title,
                    amount=currency(period.status.received),
                )
                + format_html(
                    '<a class="popup-link" href="{href}" style="background-position: 0 0" title="{title}">'
                    '<img src="{icon}" alt="+"/></a>',
                    href=reverse("admin:leprikon_subjectreceivedpayment_add") + f"?target_registration={obj.id}",
                    title=_("add payment"),
                    icon=static("admin/img/icon-addlink.svg"),
                )
            )
        return "<br/>".join(html)

    received_payments.allow_tags = True
    received_payments.short_description = _("received payments")


@admin.register(CourseDiscount)
class CourseDiscountAdmin(PdfExportAdminMixin, SubjectDiscountBaseAdmin):
    form = CourseDiscountAdminForm
    list_display = ("accounted", "registration", "subject", "period", "amount_html", "explanation")
    list_export = ("accounted", "registration", "subject", "period", "amount", "explanation")
    closed_fields = ("accounted", "registration", "period", "amount")



import colorsys
import logging
from collections import OrderedDict, namedtuple
from datetime import datetime, time
from email.mime.image import MIMEImage
from io import BytesIO
from itertools import chain
from json import loads
from os.path import basename
from tempfile import NamedTemporaryFile

import qrcode
from bankreader.models import Transaction as BankreaderTransaction
from cms.models.fields import PageField
from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.dispatch import receiver
from django.urls import reverse
from django.utils import timezone
from django.utils.encoding import force_text, smart_text
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.text import slugify
from django.utils.translation import ugettext_lazy as _
from django_pays import payment_url as pays_payment_url
from django_pays.models import Payment as PaysPayment
from djangocms_text_ckeditor.fields import HTMLField
from filer.fields.file import FilerFileField
from filer.fields.image import FilerImageField
from multiselectfield import MultiSelectField

from ..conf import settings
from ..utils import FEMALE, MALE, comma_separated, currency, lazy_paragraph as paragraph, localeconv, spayd
from .agegroup import AgeGroup
from .agreements import Agreement, AgreementOption
from .citizenship import Citizenship
from .department import Department
from .fields import BirthNumberField, ColorField, EmailField, PostalCodeField, PriceField
from .leprikonsite import LeprikonSite
from .organizations import Organization
from .pdfmail import PdfExportAndMailMixin
from .place import Place
from .printsetup import PrintSetup
from .question import Question
from .roles import Leader
from .school import School
from .schoolyear import SchoolYear
from .targetgroup import TargetGroup
from .times import AbstractTime, TimesMixin
from .transaction import AbstractTransaction, Transaction
from .utils import BankAccount, PaymentStatus, generate_variable_symbol, lazy_help_text_with_html_default

logger = logging.getLogger(__name__)

DEFAULT_TEXTS = {
    "text_registration_received": paragraph(
        _("Hello,\nthank You for submitting the registration.\nWe will inform you about its further processing.")
    ),
    "text_registration_approved": paragraph(
        _(
            "Hello,\nwe are pleased to inform You, that Your registration was approved.\n"
            "We are looking forward to see You."
        )
    ),
    "text_registration_refused": paragraph(
        _("Hello,\nwe are sorry to inform You, that Your registration was refused.")
    ),
    "text_registration_payment_request": paragraph(
        _(
            "Hello,\nwe'd like to ask You to pay for Your registration.\n"
            "If You have already made the payment recently, please ignore this message."
        )
    ),
    "text_registration_refund_offer": paragraph(
        _("Hello,\nYour registration has been overpaid.\nPlease, tell us how you wish us to refund.")
    ),
    "text_registration_canceled": paragraph(_("Hello,\nYour registration was canceled.")),
    "text_discount_granted": paragraph(_("Hello,\nwe have just grated a discount for Your registration.")),
    "text_payment_received": paragraph(
        _("Hello,\nwe have just received Your payment. Thank You.\nPlease see the recipe attached.")
    ),
    "text_payment_returned": paragraph(
        _("Hello,\nwe have just returned Your payment. Thank You.\nPlease see the recipe attached.")
    ),
}

CHAT_GROUP_BROADCAST = "B"
CHAT_GROUP_CHAT = "C"
CHAT_GROUP_TYPE_LABELS = OrderedDict(
    [
        (CHAT_GROUP_BROADCAST, _("broadcast group")),
        (CHAT_GROUP_CHAT, _("chat group")),
    ]
)

CHAT_GROUP_TYPE_HELP_TEXT = _(
    "Only the leader can write to broadcast group. Users may reply to the sender with direct messages.\n"
    "Chat group allows all members to chat with each other."
)


class SubjectType(models.Model):
    COURSE = "course"
    EVENT = "event"
    ORDERABLE = "orderable"
    subject_type_labels = OrderedDict(
        [
            (COURSE, _("course")),
            (EVENT, _("event")),
            (ORDERABLE, _("orderable event")),
        ]
    )
    subject_type_type_labels = OrderedDict(
        [
            (COURSE, _("course type")),
            (EVENT, _("event type")),
            (ORDERABLE, _("orderable event type")),
        ]
    )
    subject_type = models.CharField(_("subjects"), max_length=10, choices=subject_type_labels.items())
    name = models.CharField(_("name (singular)"), max_length=150)
    name_genitiv = models.CharField(_("name (genitiv)"), max_length=150, blank=True)
    name_akuzativ = models.CharField(_("name (akuzativ)"), max_length=150, blank=True)
    plural = models.CharField(_("name (plural)"), max_length=150)
    slug = models.SlugField()
    order = models.IntegerField(_("order"), blank=True, default=0)
    questions = models.ManyToManyField(
        Question,
        blank=True,
        related_name="+",
        verbose_name=_("additional questions"),
        help_text=_("Add additional questions to be asked in the registration form."),
    )
    registration_agreements = models.ManyToManyField(
        Agreement,
        blank=True,
        related_name="+",
        verbose_name=_("additional legal agreements"),
        help_text=_("Add additional legal agreements specific to this subject type."),
    )
    reg_print_setup = models.ForeignKey(
        PrintSetup,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("registration print setup"),
    )
    pr_print_setup = models.ForeignKey(
        PrintSetup,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("payment request print setup"),
    )
    bill_print_setup = models.ForeignKey(
        PrintSetup,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("payment print setup"),
    )
    organization = models.ForeignKey(
        Organization, blank=True, null=True, on_delete=models.SET_NULL, related_name="+", verbose_name=_("organization")
    )
    chat_group_type = models.CharField(
        _("default chat group type"),
        max_length=1,
        choices=CHAT_GROUP_TYPE_LABELS.items(),
        default=CHAT_GROUP_BROADCAST,
        help_text=CHAT_GROUP_TYPE_HELP_TEXT,
    )
    text_registration_received = HTMLField(
        _("text: registration received"),
        blank=True,
        default="",
        help_text=lazy_help_text_with_html_default("", DEFAULT_TEXTS["text_registration_received"]),
    )
    text_registration_approved = HTMLField(
        _("text: registration approved"),
        blank=True,
        default="",
        help_text=lazy_help_text_with_html_default("", DEFAULT_TEXTS["text_registration_approved"]),
    )
    text_registration_refused = HTMLField(
        _("text: registration refused"),
        blank=True,
        default="",
        help_text=lazy_help_text_with_html_default("", DEFAULT_TEXTS["text_registration_refused"]),
    )
    text_registration_payment_request = HTMLField(
        _("text: registration payment request"),
        blank=True,
        default="",
        help_text=lazy_help_text_with_html_default("", DEFAULT_TEXTS["text_registration_payment_request"]),
    )
    text_registration_refund_offer = HTMLField(
        _("text: registration refund offer"),
        blank=True,
        default="",
        help_text=lazy_help_text_with_html_default("", DEFAULT_TEXTS["text_registration_refund_offer"]),
    )
    text_registration_canceled = HTMLField(
        _("text: registration canceled"),
        blank=True,
        default="",
        help_text=lazy_help_text_with_html_default("", DEFAULT_TEXTS["text_registration_canceled"]),
    )
    text_discount_granted = HTMLField(
        _("text: discount granted"),
        blank=True,
        default="",
        help_text=lazy_help_text_with_html_default("", DEFAULT_TEXTS["text_discount_granted"]),
    )
    text_payment_received = HTMLField(
        _("text: payment received"),
        blank=True,
        default="",
        help_text=lazy_help_text_with_html_default("", DEFAULT_TEXTS["text_payment_received"]),
    )
    text_payment_returned = HTMLField(
        _("text: payment returned"),
        blank=True,
        default="",
        help_text=lazy_help_text_with_html_default("", DEFAULT_TEXTS["text_payment_returned"]),
    )

    class Meta:
        app_label = "leprikon"
        ordering = ("order",)
        verbose_name = _("subject type")
        verbose_name_plural = _("subject types")

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        if self.name_genitiv is None:
            self.name_genitiv = self.name
        if self.name_akuzativ is None:
            self.name_akuzativ = self.name
        super().save(*args, **kwargs)

    @cached_property
    def all_questions(self):
        return list(self.questions.all())

    @cached_property
    def all_attachments(self):
        return list(self.attachments.all())

    def get_absolute_url(self):
        return reverse(self.slug + ":subject_list")

    def get_journals_url(self):
        change_list = reverse("admin:leprikon_journal_changelist")
        return f"{change_list}?subject__subject_type__id__exact={self.id}"

    def get_subjects_url(self):
        change_list = reverse(f"admin:leprikon_{self.subject_type}_changelist")
        return f"{change_list}?subject_type__id__exact={self.id}"

    def get_registrations_url(self):
        change_list = reverse(f"admin:leprikon_{self.subject_type}registration_changelist")
        return f"{change_list}?subject__subject_type__id__exact={self.id}"

    def get_discounts_url(self):
        change_list = reverse(f"admin:leprikon_{self.subject_type}discount_changelist")
        return f"{change_list}?registration__subject__subject_type__id__exact={self.id}"

    def get_received_payments_url(self):
        change_list = reverse("admin:leprikon_subjectreceivedpayment_changelist")
        return f"{change_list}?target_registration__subject__subject_type__id__exact={self.id}"

    def get_returned_payments_url(self):
        change_list = reverse("admin:leprikon_subjectreturnedpayment_changelist")
        return f"{change_list}?source_registration__subject__subject_type__id__exact={self.id}"


class BaseAttachment(models.Model):
    file = FilerFileField(on_delete=models.CASCADE, related_name="+")
    public = models.BooleanField(
        _("public"), default=True, help_text=_("The attachment will be available before registration.")
    )
    events = MultiSelectField(
        _("send when"),
        blank=True,
        choices=(
            ("registration_received", _("registration received")),
            ("registration_approved", _("registration approved")),
            ("registration_refused", _("registration refused")),
            ("registration_payment_request", _("payment requested")),
            ("registration_refund_offer", _("payment requested")),
            ("registration_canceled", _("registration canceled")),
            ("discount_granted", _("discount granted")),
            ("payment_received", _("payment received")),
            ("payment_returned", _("payment returned")),
        ),
        default=[],
        help_text=_("The attachment will be sent with notification on selected events."),
    )
    order = models.IntegerField(_("order"), blank=True, default=0)

    class Meta:
        abstract = True
        app_label = "leprikon"
        ordering = ("order",)
        verbose_name = _("attachment")
        verbose_name_plural = _("attachments")

    def __str__(self):
        return force_text(self.file)


class SubjectTypeAttachment(BaseAttachment):
    subject_type = models.ForeignKey(
        SubjectType, on_delete=models.CASCADE, related_name="attachments", verbose_name=_("subject type")
    )


class SubjectGroup(models.Model):
    name = models.CharField(_("name"), max_length=150)
    plural = models.CharField(_("plural"), max_length=150)
    color = ColorField(_("color"))
    order = models.IntegerField(_("order"), blank=True, default=0)
    subject_types = models.ManyToManyField(SubjectType, related_name="groups", verbose_name=_("subject type"))

    class Meta:
        app_label = "leprikon"
        ordering = ("order",)
        verbose_name = _("subject group")
        verbose_name_plural = _("subject groups")

    def __str__(self):
        return self.name

    @cached_property
    def font_color(self):
        (h, s, v) = colorsys.rgb_to_hsv(
            int(self.color[1:3], 16) / 255.0,
            int(self.color[3:5], 16) / 255.0,
            int(self.color[5:6], 16) / 255.0,
        )
        if v > 0.5:
            v = 0
        else:
            v = 1
        if s > 0.5:
            s = 0
        else:
            s = 1
        (r, g, b) = colorsys.hsv_to_rgb(h, s, v)
        return "#{:02x}{:02x}{:02x}".format(
            int(r * 255),
            int(g * 255),
            int(b * 255),
        )


class Subject(TimesMixin, models.Model):
    PARTICIPANTS = "P"
    GROUPS = "G"
    REGISTRATION_TYPE_CHOICES = [
        (PARTICIPANTS, _("participants")),
        (GROUPS, _("groups")),
    ]
    REGISTRATION_TYPES = dict(REGISTRATION_TYPE_CHOICES)

    school_year = models.ForeignKey(
        SchoolYear, editable=False, on_delete=models.CASCADE, related_name="subjects", verbose_name=_("school year")
    )
    subject_type = models.ForeignKey(
        SubjectType, on_delete=models.PROTECT, related_name="subjects", verbose_name=_("subject type")
    )
    registration_type = models.CharField(_("registration type"), choices=REGISTRATION_TYPE_CHOICES, max_length=1)
    code = models.PositiveSmallIntegerField(_("accounting code"), blank=True, default=0)
    name = models.CharField(_("name"), max_length=150)
    description = HTMLField(_("description"), blank=True, default="")
    department = models.ForeignKey(
        Department,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="subjects",
        verbose_name=_("department"),
    )
    groups = models.ManyToManyField(SubjectGroup, verbose_name=_("groups"), related_name="subjects", blank=True)
    place = models.ForeignKey(
        Place, blank=True, null=True, on_delete=models.SET_NULL, related_name="subjects", verbose_name=_("place")
    )
    age_groups = models.ManyToManyField(AgeGroup, related_name="subjects", verbose_name=_("age groups"))
    target_groups = models.ManyToManyField(TargetGroup, related_name="subjects", verbose_name=_("target groups"))
    leaders = models.ManyToManyField(Leader, blank=True, related_name="subjects", verbose_name=_("leaders"))
    price = PriceField(_("price"), blank=True, null=True)
    public = models.BooleanField(_("public"), default=False)
    reg_from = models.DateTimeField(_("registration active from"), blank=True, null=True)
    reg_to = models.DateTimeField(_("registration active to"), blank=True, null=True)
    photo = FilerImageField(verbose_name=_("photo"), blank=True, null=True, related_name="+", on_delete=models.SET_NULL)
    page = PageField(blank=True, null=True, on_delete=models.SET_NULL, related_name="+", verbose_name=_("page"))
    min_participants_count = models.PositiveIntegerField(
        _("minimal participants count per registration"),
        default=1,
        help_text=_("Participant details include birth number (birth day), age group, contacts, parent, etc."),
    )
    max_participants_count = models.PositiveIntegerField(
        _("maximal participants count per registration"),
        default=1,
        help_text=_("Participant details include birth number (birth day), age group, contacts, parent, etc."),
    )
    min_group_members_count = models.PositiveIntegerField(
        _("minimal group members count per registration"),
        help_text=_("Group member details only include name and note."),
    )
    max_group_members_count = models.PositiveIntegerField(
        _("maximal group members count per registration"),
        help_text=_("Group member details only include name and note."),
    )
    min_registrations_count = models.PositiveIntegerField(_("minimal registrations count"), blank=True, null=True)
    max_registrations_count = models.PositiveIntegerField(_("maximal registrations count"), blank=True, null=True)
    min_due_date_days = models.PositiveIntegerField(_("minimal number of days to due date"), default=3)
    note = models.CharField(_("note"), max_length=300, blank=True, default="")
    questions = models.ManyToManyField(
        Question,
        verbose_name=_("additional questions"),
        related_name="+",
        blank=True,
        help_text=_("Add additional questions to be asked in the registration form."),
    )
    registration_agreements = models.ManyToManyField(
        Agreement,
        blank=True,
        related_name="+",
        verbose_name=_("additional legal agreements"),
        help_text=_("Add additional legal agreements specific for this subject."),
    )
    reg_print_setup = models.ForeignKey(
        PrintSetup,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("registration print setup"),
        help_text=_("Only use to set value specific for this subject."),
    )
    pr_print_setup = models.ForeignKey(
        PrintSetup,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("payment request print setup"),
        help_text=_("Only use to set value specific for this subject."),
    )
    bill_print_setup = models.ForeignKey(
        PrintSetup,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("payment print setup"),
        help_text=_("Only use to set value specific for this subject."),
    )
    organization = models.ForeignKey(
        Organization,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("organization"),
        help_text=_("Only use to set value specific for this subject."),
    )
    chat_group_type = models.CharField(
        _("chat group type"),
        blank=True,
        max_length=1,
        null=True,
        choices=CHAT_GROUP_TYPE_LABELS.items(),
        help_text=CHAT_GROUP_TYPE_HELP_TEXT,
    )
    text_registration_received = HTMLField(_("text: registration received"), blank=True, default="")
    text_registration_approved = HTMLField(_("text: registration approved"), blank=True, default="")
    text_registration_refused = HTMLField(_("text: registration refused"), blank=True, default="")
    text_registration_payment_request = HTMLField(_("text: registration payment request"), blank=True, default="")
    text_registration_refund_offer = HTMLField(_("text: registration refund offer"), blank=True, default="")
    text_registration_canceled = HTMLField(_("text: registration canceled"), blank=True, default="")
    text_discount_granted = HTMLField(_("text: discount granted"), blank=True, default="")
    text_payment_received = HTMLField(_("text: payment received"), blank=True, default="")
    text_payment_returned = HTMLField(_("text: payment returned"), blank=True, default="")

    class Meta:
        app_label = "leprikon"
        ordering = ("code", "name")
        verbose_name = _("subject")
        verbose_name_plural = _("subjects")

    def __str__(self):
        return "{} {}".format(self.school_year, self.display_name)

    def save(self, *args, **kwargs):
        if self.registration_type_participants:
            self.min_group_members_count = 0
            self.max_group_members_count = 0
        elif self.registration_type_groups:
            self.min_participants_count = 0
            self.max_participants_count = 0
        # For some reason django doesn't set the default value for code,
        # which leads to database error.
        self.code = self.code or 0
        super().save(*args, **kwargs)

    @cached_property
    def registration_type_participants(self):
        return self.registration_type == self.PARTICIPANTS

    @cached_property
    def registration_type_groups(self):
        return self.registration_type == self.GROUPS

    @cached_property
    def subject(self):
        if self.subject_type.subject_type == self.subject_type.COURSE:
            return self.course
        else:
            return self.event

    @cached_property
    def display_name(self):
        if settings.LEPRIKON_SHOW_SUBJECT_CODE and self.code:
            return "{} – {}".format(self.code, self.name)
        else:
            return self.name

    @cached_property
    def all_variants(self):
        return list(self.variants.all())

    @cached_property
    def all_groups(self):
        return list(self.groups.all())

    @cached_property
    def all_age_groups(self):
        return list(self.age_groups.all())

    @cached_property
    def all_target_groups(self):
        return list(self.target_groups.all())

    @cached_property
    def all_leaders(self):
        return list(self.leaders.all())

    @cached_property
    def all_questions(self):
        return set(self.subject_type.all_questions + list(self.questions.all()))

    @cached_property
    def all_attachments(self):
        return list(self.attachments.all())

    @cached_property
    def public_attachments(self):
        return [
            attachment for attachment in (self.subject_type.all_attachments + self.all_attachments) if attachment.public
        ]

    @cached_property
    def all_registrations(self):
        return list(self.registrations.all())

    @property
    def registration_allowed(self):
        now = timezone.now()
        return (
            self.price is not None
            and (self.reg_from or self.reg_to)
            and (self.reg_from is None or self.reg_from <= now)
            and (self.reg_to is None or self.reg_to > now)
        )

    @property
    def registration_not_allowed_message(self):
        now = timezone.now()
        if self.price is not None:
            if self.reg_from > now:
                return _("Registering will start on {}.").format(self.reg_from)
            if self.reg_to < now:
                return _("Registering ended on {}").format(self.reg_to)
        return _("Registering is currently not allowed.")

    @property
    def full(self):
        return self.max_registrations_count and self.active_registrations_count >= self.max_registrations_count

    def get_absolute_url(self):
        return reverse(self.subject_type.slug + ":subject_detail", args=(self.id,))

    def get_registration_url(self):
        return reverse(self.subject_type.slug + ":subject_registration_form", args=(self.id,))

    def get_edit_url(self):
        return reverse("admin:leprikon_{}_change".format(self.subject._meta.model_name), args=(self.id,))

    def get_age_groups_list(self):
        return comma_separated(self.all_age_groups)

    get_age_groups_list.short_description = _("age groups list")

    def get_groups_list(self):
        return comma_separated(self.all_groups)

    get_groups_list.short_description = _("groups list")

    def get_leaders_list(self):
        return comma_separated(self.all_leaders)

    get_leaders_list.short_description = _("leaders list")

    def get_target_groups_list(self):
        return comma_separated(self.all_target_groups)

    get_target_groups_list.short_description = _("target groups list")

    def get_print_setup(self, event):
        return PrintSetup()

    def get_template_variants(self):
        return (
            self.subject_type.slug,
            self.subject_type.subject_type,
            "subject",
        )

    @property
    def active_registrations(self):
        return self.registrations.filter(canceled=None)

    @cached_property
    def active_registrations_count(self):
        return self.active_registrations.count()

    @property
    def approved_registrations(self):
        return self.active_registrations.exclude(approved=None)

    @cached_property
    def approved_registrations_count(self):
        return self.approved_registrations.count()

    @cached_property
    def all_approved_registrations(self):
        return list(self.approved_registrations.all())

    @property
    def unapproved_registrations(self):
        return self.active_registrations.filter(approved=None)

    @cached_property
    def unapproved_registrations_count(self):
        return self.unapproved_registrations.count()

    @cached_property
    def all_unapproved_registrations(self):
        return list(self.unapproved_registrations.all())

    @cached_property
    def inactive_registrations_count(self):
        return self.inactive_registrations.count()

    @cached_property
    def all_inactive_registrations(self):
        return list(self.inactive_registrations.all())

    @cached_property
    def all_journals(self):
        return list(self.journals.all())

    @cached_property
    def all_registration_agreements(self):
        return sorted(
            chain(
                self.registration_agreements.all(),
                self.subject_type.registration_agreements.all(),
                LeprikonSite.objects.get_current().registration_agreements.all(),
            ),
            key=lambda agreement: agreement.order,
        )


class SubjectTime(AbstractTime):
    subject = models.ForeignKey(Subject, on_delete=models.CASCADE, related_name="times", verbose_name=_("course"))

    class Meta:
        app_label = "leprikon"
        ordering = ("day_of_week", "start")
        verbose_name = _("time")
        verbose_name_plural = _("times")


class SubjectVariant(models.Model):
    subject = models.ForeignKey(Subject, on_delete=models.CASCADE, related_name="variants", verbose_name=_("subject"))
    name = models.CharField(_("variant name"), max_length=150)
    description = HTMLField(_("variant description"), blank=True, default="")
    price = PriceField(_("price"), blank=True, null=True)
    order = models.IntegerField(_("order"), blank=True, default=0)

    class Meta:
        app_label = "leprikon"
        ordering = ("order",)
        verbose_name = _("variant")
        verbose_name_plural = _("variants")

    def __str__(self):
        return self.name

    def get_price(self):
        return self.price if self.price is not None else self.subject.price


class SubjectAttachment(BaseAttachment):
    subject = models.ForeignKey(
        Subject, on_delete=models.CASCADE, related_name="attachments", verbose_name=_("subject")
    )


class SubjectRegistration(PdfExportAndMailMixin, models.Model):
    object_name = "registration"
    slug = models.SlugField(editable=False, max_length=250, null=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        related_name="leprikon_registrations",
        verbose_name=_("user"),
    )
    subject = models.ForeignKey(
        Subject, on_delete=models.PROTECT, related_name="registrations", verbose_name=_("subject")
    )
    subject_variant = models.ForeignKey(
        SubjectVariant, null=True, on_delete=models.PROTECT, related_name="registrations", verbose_name=_("variant")
    )
    price = PriceField(_("price"), editable=False)

    created = models.DateTimeField(_("time of registration"), editable=False, auto_now_add=True)
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        null=True,
        on_delete=models.PROTECT,
        related_name="+",
        verbose_name=_("created by"),
    )
    approved = models.DateTimeField(_("time of approval"), editable=False, null=True)
    approved_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        null=True,
        on_delete=models.PROTECT,
        related_name="+",
        verbose_name=_("approved by"),
    )
    payment_requested = models.DateTimeField(_("payment request time"), editable=False, null=True)
    payment_requested_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        null=True,
        on_delete=models.PROTECT,
        related_name="+",
        verbose_name=_("payment requested by"),
    )
    refund_offered = models.DateTimeField(_("payment request time"), editable=False, null=True)
    refund_offered_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        null=True,
        on_delete=models.PROTECT,
        related_name="+",
        verbose_name=_("payment requested by"),
    )
    cancelation_requested = models.DateTimeField(_("time of cancellation request"), editable=False, null=True)
    cancelation_requested_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        null=True,
        on_delete=models.PROTECT,
        related_name="+",
        verbose_name=_("cancelation requested by"),
    )
    canceled = models.DateTimeField(_("time of cancellation"), editable=False, null=True)
    canceled_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        null=True,
        on_delete=models.PROTECT,
        related_name="+",
        verbose_name=_("canceled by"),
    )
    note = models.CharField(_("note"), max_length=300, blank=True, default="")

    questions = models.ManyToManyField(Question, editable=False, related_name="registrations")
    agreements = models.ManyToManyField(Agreement, editable=False, related_name="registrations")
    agreement_options = models.ManyToManyField(AgreementOption, blank=True, verbose_name=_("legal agreements"))

    variable_symbol = models.BigIntegerField(_("variable symbol"), db_index=True, editable=False, null=True)

    registration_link = models.ForeignKey(
        "leprikon.RegistrationLink",
        editable=False,
        null=True,
        on_delete=models.SET_NULL,
        related_name="registrations",
        verbose_name=_("registration link"),
    )

    class Meta:
        app_label = "leprikon"
        verbose_name = _("registration")
        verbose_name_plural = _("registrations")

    def __str__(self):
        return "{} - {}".format(
            self.subject,
            self.group
            if self.subject.registration_type_groups
            else comma_separated([p.full_name for p in self.all_participants]),
        )

    def get_edit_url(self):
        return reverse("admin:leprikon_{}_change".format(self.subjectregistration._meta.model_name), args=(self.id,))

    def get_payment_request_url(self):
        return reverse("leprikon:registration_payment_request", args=(self.id, self.variable_symbol))

    @cached_property
    def subjectregistration(self):
        if self.subject.subject_type.subject_type == self.subject.subject_type.COURSE:
            return self.courseregistration
        elif self.subject.subject_type.subject_type == self.subject.subject_type.EVENT:
            return self.eventregistration
        elif self.subject.subject_type.subject_type == self.subject.subject_type.ORDERABLE:
            return self.orderableregistration
        raise NotImplementedError()

    @cached_property
    def all_participants(self):
        return list(self.participants.all())

    def participants_list(self):
        return "\n".join(map(str, self.all_participants))

    participants_list.short_description = _("participants")

    def participants_list_html(self):
        return mark_safe("<br/>".join(map(str, self.all_participants)))

    participants_list_html.short_description = _("participants")

    @cached_property
    def all_group_members(self):
        return list(self.group_members.all())

    def group_members_list(self):
        return "\n".join(map(str, self.all_group_members))

    group_members_list.short_description = _("group members")

    def group_members_list_html(self):
        return mark_safe("<br/>".join(map(str, self.all_group_members)))

    group_members_list_html.short_description = _("group members")

    @cached_property
    def all_questions(self):
        return list(self.questions.all())

    @cached_property
    def all_agreements(self):
        return list(self.agreements.all())

    @cached_property
    def all_agreement_options(self):
        return list(self.agreement_options.all())

    def agreement_options_list(self):
        lines = []
        for agreement in self.all_agreements:
            lines.append(agreement.name + ":")
            for option in agreement.all_options:
                lines.append(
                    "{checkbox} {option_name}".format(
                        checkbox=("☑" if option.required or option in self.all_agreement_options else "☐"),
                        option_name=option.name,
                    )
                )
        return "\n".join(lines)

    agreement_options_list.short_description = _("agreement options")

    @cached_property
    def all_attachments(self):
        return self.subject.all_attachments + self.subject.subject_type.all_attachments

    @cached_property
    def all_recipients(self):
        recipients = set()
        if self.user.email:
            recipients.add(self.user.email)
        for participant in self.all_participants:
            recipients.update(participant.all_recipients)
        return recipients

    @cached_property
    def all_discounts(self):
        return list(self.discounts.all())

    @cached_property
    def all_received_payments(self):
        return list(self.received_payments.all())

    @cached_property
    def all_returned_payments(self):
        return list(self.returned_payments.all())

    @cached_property
    def payment_status(self) -> PaymentStatus:
        return self.get_payment_status()

    def get_payment_status(self, d=None) -> PaymentStatus:
        return self.subjectregistration.get_payment_status(d)

    @cached_property
    def organization(self):
        return (
            self.subject.organization
            or self.subject.subject_type.organization
            or LeprikonSite.objects.get_current().organization
            or Organization()
        )

    @cached_property
    def spayd(self):
        org = self.organization
        return spayd(
            ("ACC", ("%s+%s" % (org.iban, org.bic)) if org.bic else org.iban),
            ("AM", self.payment_status.amount_due),
            ("CC", localeconv["int_curr_symbol"].strip()),
            ("MSG", "%s, %s" % (self.subject.name[:29], str(self)[:29])),
            ("RN", slugify(self.organization.name).replace("*", "")[:35]),
            ("X-VS", self.variable_symbol),
        )

    @cached_property
    def payment_url(self):
        leprikon_site = LeprikonSite.objects.get_current()
        return pays_payment_url(
            gateway=leprikon_site.payment_gateway,
            order_id=self.variable_symbol,
            amount=self.payment_status.amount_due,
            email=self.user.email,
        )

    @cached_property
    def text_registration_received(self):
        return (
            self.subject.text_registration_received
            or self.subject.subject_type.text_registration_received
            or DEFAULT_TEXTS["text_registration_received"]
        )

    @cached_property
    def text_registration_approved(self):
        return (
            self.subject.text_registration_approved
            or self.subject.subject_type.text_registration_approved
            or DEFAULT_TEXTS["text_registration_approved"]
        )

    @cached_property
    def text_registration_refused(self):
        return (
            self.subject.text_registration_refused
            or self.subject.subject_type.text_registration_refused
            or DEFAULT_TEXTS["text_registration_refused"]
        )

    @cached_property
    def text_registration_payment_request(self):
        return (
            self.subject.text_registration_payment_request
            or self.subject.subject_type.text_registration_payment_request
            or DEFAULT_TEXTS["text_registration_payment_request"]
        )

    @cached_property
    def text_registration_refund_offer(self):
        return (
            self.subject.text_registration_refund_offer
            or self.subject.subject_type.text_registration_refund_offer
            or DEFAULT_TEXTS["text_registration_refund_offer"]
        )

    @cached_property
    def text_registration_canceled(self):
        return (
            self.subject.text_registration_canceled
            or self.subject.subject_type.text_registration_canceled
            or DEFAULT_TEXTS["text_registration_canceled"]
        )

    @cached_property
    def text_discount_granted(self):
        return (
            self.subject.text_discount_granted
            or self.subject.subject_type.text_discount_granted
            or DEFAULT_TEXTS["text_discount_granted"]
        )

    @cached_property
    def text_payment_received(self):
        return (
            self.subject.text_payment_received
            or self.subject.subject_type.text_payment_received
            or DEFAULT_TEXTS["text_payment_received"]
        )

    @cached_property
    def text_payment_returned(self):
        return (
            self.subject.text_payment_returned
            or self.subject.subject_type.text_payment_returned
            or DEFAULT_TEXTS["text_payment_returned"]
        )

    def get_discounts(self, d):
        if d:
            return [p for p in self.all_discounts if p.accounted.date() <= d]
        else:
            return self.all_discounts

    def get_received_payments(self, d):
        if d:
            return [p for p in self.all_received_payments if p.accounted.date() <= d]
        else:
            return self.all_received_payments

    def get_returned_payments(self, d):
        if d:
            return [r for r in self.all_returned_payments if r.accounted.date() <= d]
        else:
            return self.all_returned_payments

    def get_pdf_filename(self, event):
        if event == "payment_request":
            return basename(self.get_payment_request_url())
        return self.slug + ".pdf"

    def get_discounted(self, d=None):
        return sum(p.amount for p in self.get_discounts(d))

    def get_received(self, d=None):
        return sum(p.amount for p in self.get_received_payments(d))

    def get_returned(self, d=None):
        return sum(r.amount for r in self.get_returned_payments(d))

    def get_qr_code(self):
        output = BytesIO()
        self.write_qr_code(output)
        output.seek(0)
        return output.read()

    def write_qr_code(self, output):
        qrcode.make(self.spayd, border=1).save(output)

    def write_pdf(self, event, output):
        if event == "payment_request":
            with NamedTemporaryFile(buffering=0, suffix=".png") as qr_code_file:
                self.write_qr_code(qr_code_file)
                qr_code_file.flush()
                self.qr_code_filename = qr_code_file.name
                return super().write_pdf(event, output)
        else:
            return super().write_pdf(event, output)

    @transaction.atomic
    def approve(self, approved_by):
        if self.approved is None:
            self.canceled = None
            self.canceled_by = None
            self.approved = timezone.now()
            self.approved_by = approved_by
            self.save()
            self.send_mail("approved")
            if not self.payment_requested:
                self.request_payment(approved_by)
            if len(self.subject.all_journals) == 1:
                journal = self.subject.all_journals[0]
                for participant in self.all_participants:
                    journal.participants.add(participant)
        else:
            raise ValidationError(
                (
                    _("Unfortunately, it is not possible to restore canceled registration {r}. Please create new one.")
                    if self.canceled
                    else _("The registration {r} has already been approved.")
                ).format(r=self)
            )

    @transaction.atomic
    def request_payment(self, payment_requested_by):
        if not self.payment_requested:
            self.payment_requested = timezone.now()
            self.payment_requested_by = payment_requested_by
            self.save()
            # refresh all cached statuses
            self = type(self).objects.get(id=self.id)
        if self.payment_status.amount_due:
            self.send_mail("payment_request")

    @transaction.atomic
    def offer_refund(self, refund_offered_by):
        if self.payment_status.overpaid:
            self.refund_offered = timezone.now()
            self.refund_offered_by = refund_offered_by
            self.save()
            self.send_mail("refund_offer")

    @transaction.atomic
    def generate_refund_request(self, generated_by):
        from .refundrequest import RefundRequest

        if not self.payment_status.overpaid:
            return
        remote_accounts = set(
            account
            for account in self.received_payments.annotate(
                account=models.F("bankreader_transaction__remote_account_number")
            ).values_list("account", flat=True)
            if account
        )
        if len(remote_accounts) == 1:
            RefundRequest.objects.create(
                registration=self,
                requested_by=generated_by,
                bank_account=BankAccount(remote_accounts.pop()),
            )

    def refuse(self, refused_by):
        if self.approved is None:
            if self.canceled is None:
                with transaction.atomic():
                    self.canceled = timezone.now()
                    self.canceled_by = refused_by
                    self.save()
                    self.send_mail("refused")
            else:
                raise ValidationError(_("The registration {r} has already been refued.").format(r=self))
        else:
            raise ValidationError(
                _(
                    "Unfortunately, it is not possible to refuse the registration {r}. However, You may cancel it."
                ).format(r=self)
            )

    def cancel(self, canceled_by):
        if self.approved:
            if self.canceled is None:
                with transaction.atomic():
                    self.canceled = timezone.now()
                    self.canceled_by = canceled_by
                    self.save()
                    self.send_mail("canceled")
            else:
                raise ValidationError(_("The registration {r} has already been canceled.").format(r=self))
        else:
            raise ValidationError(
                _(
                    "Unfortunately, it is not possible to cancel the registration {r}. However, You may refuse it."
                ).format(r=self)
            )

    def generate_variable_symbol_and_slug(self):
        self.variable_symbol = generate_variable_symbol(self)
        self.slug = "{}-{}".format(
            slugify(
                "{}-{}".format(
                    self.subject.name[:100],
                    self.group
                    if self.subject.registration_type_groups
                    else comma_separated([p.full_name for p in self.all_participants]),
                )
            )[:200],
            self.variable_symbol,
        )
        self.save()

    def get_print_setup(self, event):
        if event == "payment_request":
            return (
                self.subject.pr_print_setup
                or self.subject.subject_type.pr_print_setup
                or LeprikonSite.objects.get_current().pr_print_setup
                or PrintSetup()
            )
        return (
            self.subject.reg_print_setup
            or self.subject.subject_type.reg_print_setup
            or LeprikonSite.objects.get_current().reg_print_setup
            or PrintSetup()
        )

    def get_template_variants(self):
        return (
            self.subject.subject_type.slug,
            self.subject.subject_type.subject_type,
            "subject",
        )

    def get_attachments(self, event):
        attachments = []

        if event == "payment_request":
            if self.organization.iban:
                qr_code = MIMEImage(self.get_qr_code())
                qr_code.add_header("Content-ID", "<qr_code>")
                attachments.append(qr_code)
            attachments.append(self.get_pdf_attachment(event))

        attachments += [
            (basename(attachment.file.file.path), open(attachment.file.file.path, "rb").read())
            for attachment in self.all_attachments
            if f"registration_{event}" in attachment.events
        ]
        return attachments or None


class QuestionsMixin:
    def get_answers(self):
        return loads(self.answers)

    def get_questions_and_answers(self):
        answers = self.get_answers()
        for q in self.registration.all_questions:
            yield {
                "question": q.question,
                "answer": answers.get(q.name, ""),
            }


class PersonMixin:
    @cached_property
    def full_name(self):
        return "{} {}".format(self.first_name, self.last_name)

    @cached_property
    def address(self):
        return "{}, {}, {}".format(self.street, self.city, self.postal_code)

    @cached_property
    def contact(self):
        if self.email and self.phone:
            return "{}, {}".format(self.phone, self.email)
        else:
            return self.email or self.phone or ""


class SchoolMixin:
    @cached_property
    def school_name(self):
        return self.school and smart_text(self.school) or self.school_other

    @cached_property
    def school_and_class(self):
        return "{}, {}".format(self.school_name, self.school_class)


class SubjectRegistrationParticipant(SchoolMixin, PersonMixin, QuestionsMixin, models.Model):
    registration = models.ForeignKey(
        SubjectRegistration, on_delete=models.CASCADE, related_name="participants", verbose_name=_("registration")
    )
    first_name = models.CharField(_("first name"), max_length=30)
    last_name = models.CharField(_("last name"), max_length=30)
    citizenship = models.ForeignKey(
        Citizenship, on_delete=models.PROTECT, related_name="+", verbose_name=_("citizenship")
    )
    birth_num = BirthNumberField(_("birth number"), blank=True, null=True)
    birth_date = models.DateField(_("birth date"))
    gender = models.CharField(
        _("gender"), max_length=1, choices=((MALE, _("male / boy")), (FEMALE, _("female / girl")))
    )
    age_group = models.ForeignKey(AgeGroup, on_delete=models.PROTECT, related_name="+", verbose_name=_("age group"))
    street = models.CharField(_("street"), max_length=150)
    city = models.CharField(_("city"), max_length=150)
    postal_code = PostalCodeField(_("postal code"))
    phone = models.CharField(_("phone"), max_length=30, blank=True, default="")
    email = EmailField(_("email address"), blank=True, default="")

    school = models.ForeignKey(
        School, blank=True, null=True, on_delete=models.PROTECT, related_name="+", verbose_name=_("school")
    )
    school_other = models.CharField(_("other school"), max_length=150, blank=True, default="")
    school_class = models.CharField(_("class"), max_length=30, blank=True, default="")
    health = models.TextField(_("health"), blank=True, default="")

    has_parent1 = models.BooleanField(_("first parent"), default=False)
    parent1_first_name = models.CharField(_("first name"), max_length=30, blank=True, null=True)
    parent1_last_name = models.CharField(_("last name"), max_length=30, blank=True, null=True)
    parent1_street = models.CharField(_("street"), max_length=150, blank=True, null=True)
    parent1_city = models.CharField(_("city"), max_length=150, blank=True, null=True)
    parent1_postal_code = PostalCodeField(_("postal code"), blank=True, null=True)
    parent1_phone = models.CharField(_("phone"), max_length=30, blank=True, null=True)
    parent1_email = EmailField(_("email address"), blank=True, null=True)

    has_parent2 = models.BooleanField(_("second parent"), default=False)
    parent2_first_name = models.CharField(_("first name"), max_length=30, blank=True, null=True)
    parent2_last_name = models.CharField(_("last name"), max_length=30, blank=True, null=True)
    parent2_street = models.CharField(_("street"), max_length=150, blank=True, null=True)
    parent2_city = models.CharField(_("city"), max_length=150, blank=True, null=True)
    parent2_postal_code = PostalCodeField(_("postal code"), blank=True, null=True)
    parent2_phone = models.CharField(_("phone"), max_length=30, blank=True, null=True)
    parent2_email = EmailField(_("email address"), blank=True, null=True)

    answers = models.TextField(_("additional answers"), blank=True, default="{}", editable=False)

    class Meta:
        app_label = "leprikon"
        ordering = ("last_name", "first_name")
        verbose_name = _("participant")
        verbose_name_plural = _("participants")

    def __str__(self):
        return "{full_name} ({birth_year})".format(
            full_name=self.full_name,
            birth_year=self.birth_date.year,
        )

    @cached_property
    def parents(self):
        return list(p for p in [self.parent1, self.parent2] if p is not None)

    @cached_property
    def parent1(self):
        if self.has_parent1:
            return self.Parent(self, "parent1")
        else:
            return None

    parent1.short_description = _("first parent")

    @cached_property
    def parent2(self):
        if self.has_parent2:
            return self.Parent(self, "parent2")
        else:
            return None

    parent2.short_description = _("second parent")

    @cached_property
    def all_recipients(self):
        # don't use unverified emails
        # TODO: implement verification process
        return set()
        # recipients = set(parent.email for parent in self.parents if parent.email)
        # if self.email:
        #     recipients.add(self.email)
        # return recipients

    PresenceRecord = namedtuple("PresenceRecord", ("entry", "present"))

    @property
    def presences(self):
        from .journals import JournalEntry

        present_entry_ids = set(self.journal_entries.values_list("id", flat=True))

        return [
            self.PresenceRecord(entry, entry.id in present_entry_ids)
            for entry in JournalEntry.objects.filter(journal__participants=self)
        ]

    def save(self, *args, **kwargs):
        if not self.has_parent1:
            if self.has_parent2:
                self.parent1_first_name = self.parent2_first_name
                self.parent1_last_name = self.parent2_last_name
                self.parent1_street = self.parent2_street
                self.parent1_city = self.parent2_city
                self.parent1_postal_code = self.parent2_postal_code
                self.parent1_phone = self.parent2_phone
                self.parent1_email = self.parent2_email
                self.has_parent1 = True
                self.has_parent2 = False
            else:
                self.parent1_first_name = None
                self.parent1_last_name = None
                self.parent1_street = None
                self.parent1_city = None
                self.parent1_postal_code = None
                self.parent1_phone = None
                self.parent1_email = None
        if not self.has_parent2:
            self.parent2_first_name = None
            self.parent2_last_name = None
            self.parent2_street = None
            self.parent2_city = None
            self.parent2_postal_code = None
            self.parent2_phone = None
            self.parent2_email = None
        super().save(*args, **kwargs)

    class Parent(PersonMixin):
        def __init__(self, registration, role):
            self._registration = registration
            self._role = role

        def __getattr__(self, attr):
            return getattr(self._registration, "{}_{}".format(self._role, attr))

        def __str__(self):
            return self.full_name


class SubjectRegistrationGroup(SchoolMixin, PersonMixin, QuestionsMixin, models.Model):
    registration = models.OneToOneField(
        SubjectRegistration, on_delete=models.CASCADE, related_name="group", verbose_name=_("registration")
    )
    target_group = models.ForeignKey(
        TargetGroup, on_delete=models.PROTECT, related_name="+", verbose_name=_("target group")
    )
    name = models.CharField(_("group name"), blank=True, default="", max_length=150)
    first_name = models.CharField(_("first name"), max_length=30)
    last_name = models.CharField(_("last name"), max_length=30)
    street = models.CharField(_("street"), max_length=150)
    city = models.CharField(_("city"), max_length=150)
    postal_code = PostalCodeField(_("postal code"))
    phone = models.CharField(_("phone"), max_length=30)
    email = EmailField(_("email address"))
    school = models.ForeignKey(
        School, blank=True, null=True, on_delete=models.PROTECT, related_name="+", verbose_name=_("school")
    )
    school_other = models.CharField(_("other school"), max_length=150, blank=True, default="")
    school_class = models.CharField(_("class"), max_length=30, blank=True, default="")

    answers = models.TextField(_("additional answers"), blank=True, default="{}", editable=False)

    class Meta:
        app_label = "leprikon"
        verbose_name = _("registered group")
        verbose_name_plural = _("registered groups")

    def __str__(self):
        return self.name or self.full_name


class SubjectRegistrationGroupMember(models.Model):
    registration = models.ForeignKey(
        SubjectRegistration, on_delete=models.CASCADE, related_name="group_members", verbose_name=_("registration")
    )
    first_name = models.CharField(_("first name"), max_length=30)
    last_name = models.CharField(_("last name"), max_length=30)
    note = models.CharField(_("note"), max_length=150, blank=True, default="")

    class Meta:
        app_label = "leprikon"
        verbose_name = _("group member")
        verbose_name_plural = _("group members")

    def __str__(self):
        return (
            "{full_name} ({note})".format(
                full_name=self.full_name,
                note=self.note,
            )
            if self.note
            else self.full_name
        )

    @cached_property
    def full_name(self):
        return "{} {}".format(self.first_name, self.last_name)


class SubjectRegistrationBillingInfo(models.Model):
    registration = models.OneToOneField(
        SubjectRegistration, on_delete=models.CASCADE, related_name="billing_info", verbose_name=_("registration")
    )
    name = models.CharField(_("name"), max_length=150)
    street = models.CharField(_("street"), max_length=150, blank=True, default="")
    city = models.CharField(_("city"), max_length=150, blank=True, default="")
    postal_code = PostalCodeField(_("postal code"), blank=True, default="")
    company_num = models.CharField(_("company number"), max_length=8, blank=True, default="")
    vat_number = models.CharField(_("VAT number"), max_length=12, blank=True, default="")
    contact_person = models.CharField(_("contact person"), max_length=60, blank=True, default="")
    phone = models.CharField(_("phone"), max_length=30, blank=True, default="")
    email = EmailField(_("email address"), blank=True, default="")
    employee = models.CharField(_("employee ID"), max_length=150, blank=True, default="")

    class Meta:
        app_label = "leprikon"
        verbose_name = _("billing information")
        verbose_name_plural = _("billing information")

    def __str__(self):
        return self.name

    @cached_property
    def address(self):
        return ", ".join(filter(bool, (self.street, self.city, self.postal_code)))

    address.short_description = _("address")


class SubjectDiscount(AbstractTransaction):
    amount = PriceField(_("discount"), default=0)
    explanation = models.CharField(_("discount explanation"), max_length=250, blank=True, default="")

    class Meta:
        abstract = True
        app_label = "leprikon"
        verbose_name = _("discount")
        verbose_name_plural = _("discounts")
        ordering = ("accounted",)

    def __str__(self):
        return currency(self.amount)


class SubjectPaymentMixin:
    @cached_property
    def subject_organization(self):
        return (
            self.registration.subject.organization
            or self.registration.subject.subject_type.organization
            or LeprikonSite.objects.get_current().organization
            or Organization()
        )

    @cached_property
    def slug(self):
        return f"{self.registration.slug}-{self.id}"

    def get_print_setup(self, event):
        return (
            self.registration.subject.bill_print_setup
            or self.registration.subject.subject_type.bill_print_setup
            or LeprikonSite.objects.get_current().bill_print_setup
            or PrintSetup()
        )

    def get_template_variants(self):
        return (
            self.registration.subject.subject_type.slug,
            self.registration.subject.subject_type.subject_type,
            "subject",
        )

    def get_attachments(self, event):
        attachments = []

        if event == "received":
            attachments.append(self.get_pdf_attachment("pdf"))

        attachments += [
            (basename(attachment.file.file.path), open(attachment.file.file.path, "rb").read())
            for attachment in self.registration.all_attachments
            if f"{self.object_name}_{event}" in attachment.events
        ]
        return attachments or None

    @cached_property
    def all_recipients(self):
        return self.registration.all_recipients


class SubjectPayment(Transaction):
    object_name = "payment"
    transaction_types = Transaction.SUBJECTS

    class Meta:
        proxy = True
        verbose_name = _("payment")
        verbose_name_plural = _("payments")

    @cached_property
    def sub_payments(self):
        sub_payments = []
        if self.target_registration:
            sub_payments.append(
                SubjectReceivedPayment(**{key: value for key, value in self.__dict__.items() if key[0] != "_"})
            )
        if self.source_registration:
            sub_payments.append(
                SubjectReturnedPayment(**{key: value for key, value in self.__dict__.items() if key[0] != "_"})
            )
        return sub_payments


class SubjectReceivedPayment(SubjectPaymentMixin, Transaction):
    object_name = "received_payment"
    transaction_types = Transaction.PAYMENTS

    class Meta:
        proxy = True
        verbose_name = _("received payment")
        verbose_name_plural = _("received payments")

    @property
    def real_amount(self):
        return self.amount

    @cached_property
    def registration(self):
        return self.target_registration

    registration.short_description = _("target registration")


class SubjectReturnedPayment(SubjectPaymentMixin, Transaction):
    object_name = "returned_payment"
    transaction_types = Transaction.RETURNS

    class Meta:
        proxy = True
        verbose_name = _("returned payment")
        verbose_name_plural = _("returned payments")

    @property
    def real_amount(self):
        return -self.amount

    @cached_property
    def registration(self):
        return self.source_registration

    registration.short_description = _("source registration")


@receiver(models.signals.post_save, sender=PaysPayment)
def payment_create_subject_payment(instance, **kwargs):
    payment = instance
    # check realized payment
    if payment.status != PaysPayment.REALIZED:
        return
    # check registration
    try:
        registration = SubjectRegistration.objects.get(variable_symbol=int(payment.order_id))
    except (ValueError, SubjectRegistration.DoesNotExist):
        return
    # create payment
    Transaction.objects.create(
        transaction_type=Transaction.PAYMENT_ONLINE,
        target_registration=registration,
        accounted=payment.created,
        amount=payment.amount / payment.base_units,
        note=_("received online payment"),
        pays_payment=payment,
    )


@receiver(models.signals.post_save, sender=BankreaderTransaction)
def transaction_create_subject_payment(instance, **kwargs):
    transaction = instance
    # check variable symbol
    if not transaction.variable_symbol:
        return
    # check closure date (use closure date from cached leprikon site)
    max_closure_date = LeprikonSite.objects.get_current().max_closure_date
    if max_closure_date and transaction.accounted_date <= max_closure_date:
        return
    # check registration
    registration = SubjectRegistration.objects.filter(variable_symbol=transaction.variable_symbol).first()
    if not registration:
        return
    # create payment
    if transaction.amount < 0:
        kwargs = {
            "transaction_type": Transaction.RETURN_BANK,
            "source_registration": registration,
            "amount": -transaction.amount,
        }
    else:
        kwargs = {
            "transaction_type": Transaction.PAYMENT_BANK,
            "target_registration": registration,
            "amount": transaction.amount,
        }
    Transaction.objects.create(
        accounted=timezone.make_aware(datetime.combine(transaction.accounted_date, time(12))),
        note=_("imported from account statement"),
        bankreader_transaction=transaction,
        **kwargs,
    )



from datetime import date

from django import forms
from django.core.exceptions import ValidationError
from django.forms.models import inlineformset_factory
from django.utils.translation import ugettext_lazy as _, ungettext_lazy as ungettext

from ..models.journals import Journal, JournalEntry, JournalLeaderEntry, JournalTime
from ..models.roles import Leader
from ..models.subjects import SubjectRegistrationParticipant
from ..models.timesheets import Timesheet, TimesheetPeriod
from ..utils import comma_separated, first_upper
from .fields import ReadonlyField
from .form import FormMixin
from .widgets import CheckboxSelectMultipleBootstrap


class JournalAdminForm(forms.ModelForm):
    class Meta:
        model = Journal
        exclude = []

    def __init__(self, data=None, *args, **kwargs):
        super().__init__(data, *args, **kwargs)

        # limit choices of leaders
        leaders_choices = self.fields["leaders"].widget.choices
        leaders_choices.queryset = leaders_choices.queryset.filter(subjects=self.subject)

        participant_ids = set(
            SubjectRegistrationParticipant.objects.filter(
                registration__approved__isnull=False,
                registration__subject_id=self.subject.id,
            ).values_list("id", flat=True)
        )
        if self.instance.id:
            participant_ids |= set(self.instance.participants.values_list("id", flat=True))
            participant_ids |= set(
                SubjectRegistrationParticipant.objects.filter(
                    journal_entries__journal=self.instance,
                ).values_list("id", flat=True),
            )
        self.fields["participants"].widget.choices.queryset = SubjectRegistrationParticipant.objects.filter(
            id__in=participant_ids
        )


class JournalForm(FormMixin, forms.ModelForm):
    JournalTimeFormset = inlineformset_factory(
        Journal, JournalTime, fields=["day_of_week", "start", "end"], can_delete=True, extra=1
    )

    class Meta:
        model = Journal
        exclude = ["subject"]
        widgets = {
            "leaders": CheckboxSelectMultipleBootstrap(),
            "participants": CheckboxSelectMultipleBootstrap(),
        }

    def __init__(self, *args, **kwargs):
        self.subject = kwargs.pop("subject", None) or kwargs["instance"].subject
        super().__init__(*args, **kwargs)
        self.instance.subject = self.subject
        self.readonly_fields = [
            ReadonlyField(label=first_upper(self.subject.subject_type.name), value=self.subject),
        ]
        self.fields["leaders"].widget.choices = tuple((leader.id, leader) for leader in self.subject.all_leaders)

        participant_ids = set(
            SubjectRegistrationParticipant.objects.filter(
                registration__approved__isnull=False,
                registration__subject_id=self.subject.id,
            ).values_list("id", flat=True)
        )
        if self.instance.id:
            participant_ids |= set(self.instance.participants.values_list("id", flat=True))
            participant_ids |= set(
                SubjectRegistrationParticipant.objects.filter(
                    journal_entries__journal=self.instance,
                ).values_list("id", flat=True),
            )
        self.fields["participants"].widget.choices = tuple(
            (participant.id, participant)
            for participant in SubjectRegistrationParticipant.objects.filter(id__in=participant_ids)
        )
        self.inline_formsets = [self.JournalTimeFormset(data=kwargs.get("data"), instance=kwargs.get("instance"))]

    def is_valid(self) -> bool:
        return all((super().is_valid(), *(inline_formset.is_valid() for inline_formset in self.inline_formsets)))

    def save(self, commit=True):
        obj = super().save(commit=commit)
        for inline_formset in self.inline_formsets:
            inline_formset.instance = obj
            inline_formset.save(commit=commit)
        return obj


class JournalLeaderEntryAdminForm(forms.ModelForm):
    class Meta:
        model = JournalLeaderEntry
        fields = ["start", "end"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance.timesheet.submitted:
            try:
                del self.fields["start"]
                del self.fields["end"]
            except KeyError:
                pass

    def clean(self):
        cleaned_data = super().clean()

        # readonly start end
        if self.instance.id and self.instance.timesheet.submitted:
            cleaned_data["start"] = self.instance.start
            cleaned_data["end"] = self.instance.end

        # check entry start and end
        errors = {}
        journal_entry = self.instance.journal_entry
        if cleaned_data["start"] < journal_entry.start:
            errors["start"] = _("The journal entry starts at {start}").format(
                start=journal_entry.start,
            )
        if cleaned_data["end"] > journal_entry.end:
            errors["end"] = _("The journal entry ends at {end}").format(
                end=journal_entry.end,
            )
        if errors:
            raise ValidationError(errors)

        return cleaned_data


class JournalLeaderEntryForm(FormMixin, JournalLeaderEntryAdminForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.readonly_fields = [
            ReadonlyField(label=_("Journal"), value=self.instance.journal_entry.journal),
            ReadonlyField(label=_("Date"), value=self.instance.journal_entry.date),
            ReadonlyField(label=_("Leader"), value=self.instance.timesheet.leader),
        ]
        if self.instance.timesheet.submitted:
            self.readonly_fields += [
                ReadonlyField(label=_("Start"), value=self.instance.start),
                ReadonlyField(label=_("End"), value=self.instance.end),
            ]


class JournalEntryAdminForm(forms.ModelForm):
    class Meta:
        model = JournalEntry
        fields = ["date", "start", "end", "agenda", "participants", "participants_instructed"]

    def __init__(self, *args, **kwargs):
        self.journal = kwargs.pop("journal", None) or kwargs["instance"].journal
        super().__init__(*args, **kwargs)
        self.instance.journal = self.journal

        if self.instance.id:
            d = self.instance.date
            self.readonly_fields = [
                ReadonlyField(label=_("Date"), value=d),
            ]
            try:
                del self.fields["date"]
            except KeyError:
                pass
        else:
            last = self.instance.journal.journal_entries.last()
            if last:
                last_end = last.datetime_end or last.date
            else:
                last_end = None
            next_time = self.instance.journal.get_next_time(last_end)
            if next_time:
                self.initial["date"] = next_time.date
                self.initial["start"] = next_time.start
                self.initial["end"] = next_time.end
            else:
                self.initial["date"] = date.today()
            try:
                d = self.fields["date"].clean(kwargs["data"]["date"])
            except (KeyError, TypeError, ValidationError):
                d = self.initial["date"]
        self.fields["participants"].widget.choices.queryset = self.instance.journal.get_valid_participants(d)
        self.fields["participants"].help_text = None

        self.fields["participants_instructed"].widget.choices.queryset = self.instance.journal.get_valid_participants(d)
        self.fields["participants_instructed"].help_text = None

    def clean(self):
        date = self.cleaned_data.get("date", self.instance.date)
        start = self.cleaned_data.get("start")
        end = self.cleaned_data.get("end")

        # start and end must be both set or both None
        if start is None and end is not None:
            raise ValidationError({"start": _("If You fill in start time, You must fill end time too.")})
        elif end is None and start is not None:
            raise ValidationError({"end": _("If You fill in end time, You must fill start time too.")})

        # check overlaping entries
        if start:
            qs = JournalEntry.objects.filter(
                journal=self.instance.journal,
                date=self.cleaned_data.get("date", self.instance.date),
                start__lt=end,
                end__gt=start,
            )
            if self.instance.id:
                qs = qs.exclude(id=self.instance.id)
            if qs.exists():
                raise ValidationError(_("An overlaping entry has already been added in the journal."))

        # check instructed participants
        participants = set(self.cleaned_data.get("participants", []))
        participants_instructed = set(self.cleaned_data.get("participants_instructed", []))
        not_present_participants_instructed = participants_instructed - participants
        if not_present_participants_instructed:
            self.add_error(
                "participants_instructed",
                _("Following participants were selected as instructed, but not selected as present: {}").format(
                    comma_separated(map(str, not_present_participants_instructed))
                ),
            )
        not_instructed_participants = []
        for participant in participants - participants_instructed:
            if not participant.instructed.filter(date__lt=date).exists():
                not_instructed_participants.append(participant)
        if not_instructed_participants:
            self.add_error(
                "participants",
                _("Following participants have not yet been instructed: {}").format(
                    comma_separated(map(str, not_instructed_participants))
                ),
            )

        # check submitted leader entries
        submitted_leader_entries = [e for e in self.instance.all_leader_entries if e.timesheet.submitted]
        if submitted_leader_entries:
            max_start = min(e.start for e in submitted_leader_entries)
            min_end = max(e.end for e in submitted_leader_entries)
            errors = {}
            if (start is None) or (start > max_start):
                errors["start"] = _("Some submitted timesheet entries start at {start}").format(
                    start=max_start,
                )
            if (end is None) or (end < min_end):
                errors["end"] = _("Some submitted timesheet entries end at {end}").format(
                    end=min_end,
                )
            if errors:
                raise ValidationError(errors)

        return self.cleaned_data


class JournalEntryForm(FormMixin, JournalEntryAdminForm):
    leaders = forms.ModelMultipleChoiceField(Leader.objects, label=_("Leaders"), required=False)
    alternates = forms.ModelMultipleChoiceField(Leader.objects, label=_("Alternates"), required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.readonly_fields = [
            ReadonlyField(label=_("Journal"), value=self.journal),
        ]

        # only allow to select leaders or alterates with not submitted timesheets
        leaders = self.instance.journal.all_leaders
        alternates = [leader for leader in Leader.objects.all() if leader not in leaders]

        self.fields["leaders"].widget.choices = tuple((leader.id, leader) for leader in leaders)
        self.fields["leaders"].help_text = None
        self.fields["alternates"].widget.choices = tuple((leader.id, leader) for leader in alternates)
        self.fields["alternates"].help_text = None

        if self.instance.id:
            self.initial["leaders"] = [leader.id for leader in self.instance.all_leaders]
            self.initial["alternates"] = [leader.id for leader in self.instance.all_alternates]
        else:
            self.initial["leaders"] = [leader.id for leader in leaders]

    def clean(self):
        self.cleaned_data = super().clean()
        self.cleaned_entries = []
        self.deleted_entries = []

        d = self.cleaned_data.get("date", self.instance.date)
        if d is None:
            # no other validation makes sense without date
            return self.cleaned_data

        if (
            "start" in self.cleaned_data
            and "end" in self.cleaned_data
            and self.cleaned_data["start"]
            and self.cleaned_data["end"]
        ):
            errors = {}
            entries_by_leader = {
                "leaders": dict(
                    (entry.timesheet.leader, entry)
                    for entry in self.instance.all_leader_entries
                    if entry.timesheet.leader in self.instance.all_leaders
                ),
                "alternates": dict(
                    (entry.timesheet.leader, entry)
                    for entry in self.instance.all_leader_entries
                    if entry.timesheet.leader in self.instance.all_alternates
                ),
            }
            leaders_with_submitted_timesheets = {}
            period = TimesheetPeriod.objects.for_date(d)

            for group in ("leaders", "alternates"):
                leaders_with_submitted_timesheets[group] = []
                for leader in self.cleaned_data[group]:
                    if leader not in entries_by_leader[group]:
                        # try to create new leader entry
                        timesheet = Timesheet.objects.for_leader_and_date(leader=leader, date=d)
                        if timesheet.submitted:
                            # can not create entry
                            leaders_with_submitted_timesheets[group].append(leader)
                            continue
                        entry = JournalLeaderEntry()
                        entry.journal_entry = self.instance
                        entry.timesheet = timesheet
                        entry.start = self.cleaned_data["start"]
                        entry.end = self.cleaned_data["end"]
                    else:
                        # try to update existing leader entry
                        entry = entries_by_leader[group].pop(leader)
                        if not entry.timesheet.submitted:
                            if self.cleaned_data["start"] != self.instance.start:
                                if entry.start == self.instance.start:
                                    entry.start = self.cleaned_data["start"]
                                else:
                                    entry.start = max(entry.start, self.cleaned_data["start"])
                            if self.cleaned_data["end"] != self.instance.end:
                                if entry.end == self.instance.end:
                                    entry.end = self.cleaned_data["end"]
                                else:
                                    entry.end = min(entry.end, self.cleaned_data["end"])
                    # store cleaned entry, or delete it, if it is broken by the update
                    if entry.start < entry.end:
                        self.cleaned_entries.append(entry)
                    elif entry.id:
                        self.deleted_entries.append(entry)
                # try to delete stale entries
                for entry in entries_by_leader[group].values():
                    if entry.timesheet.submitted:
                        # can not delete entry
                        leaders_with_submitted_timesheets[group].append(entry.timesheet.leader)
                        continue
                    # store deleted entry
                    self.deleted_entries.append(entry)

            if leaders_with_submitted_timesheets["leaders"]:
                errors["leaders"] = ungettext(
                    "Leader {leaders} has already submitted timesheet for {period}.",
                    "Leaders {leaders} have already submitted timesheet for {period}.",
                    len(leaders_with_submitted_timesheets["leaders"]),
                ).format(
                    leaders=comma_separated(leaders_with_submitted_timesheets["leaders"]),
                    period=period.name,
                )
            if leaders_with_submitted_timesheets["alternates"]:
                errors["alternates"] = ungettext(
                    "Alternate {leaders} has already submitted timesheet for {period}.",
                    "Alternates {leaders} have already submitted timesheet for {period}.",
                    len(leaders_with_submitted_timesheets["alternates"]),
                ).format(
                    leaders=comma_separated(leaders_with_submitted_timesheets["alternates"]),
                    period=period.name,
                )
            if errors:
                raise ValidationError(errors)

        return self.cleaned_data

    def save(self, commit=True):
        self.instance = super().save(commit)
        for entry in self.cleaned_entries:
            entry.journal_entry = self.instance
            entry.save()
        for entry in self.deleted_entries:
            entry.delete()



from collections import namedtuple
from datetime import date, datetime, timedelta

from cms.models import CMSPlugin
from django.db import models
from django.utils.formats import date_format, time_format
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from ..conf import settings
from .agegroup import AgeGroup
from .department import Department
from .roles import Leader
from .schoolyear import SchoolYear
from .subjects import Subject, SubjectDiscount, SubjectGroup, SubjectRegistration, SubjectType
from .targetgroup import TargetGroup
from .utils import PaymentStatus


class Orderable(Subject):
    duration = models.DurationField(_("duration"), help_text=_("Format: HH:MM:SS"))
    due_from_days = models.IntegerField(
        _("number of days to send the payment request before event date"),
        blank=True,
        null=True,
        help_text=_(
            "If set, payment request will be sent this number of days before event date. "
            "If not set, payment request will be sent when registration is approved."
        ),
    )
    due_date_days = models.IntegerField(_("number of days to due date before event date"), default=0)

    class Meta:
        app_label = "leprikon"
        ordering = ("code", "name")
        verbose_name = _("orderable event")
        verbose_name_plural = _("orderable events")

    @property
    def inactive_registrations(self):
        return self.registrations.filter(canceled__isnull=False)

    def get_times_list(self):
        return self.duration

    get_times_list.short_description = _("duration")

    def copy_to_school_year(old, school_year):
        new = Orderable.objects.get(id=old.id)
        new.id, new.pk = None, None
        new.school_year = school_year
        new.public = False
        new.evaluation = ""
        new.note = ""
        new.save()
        new.groups.set(old.groups.all())
        new.age_groups.set(old.age_groups.all())
        new.target_groups.set(old.target_groups.all())
        for leader in old.all_leaders:
            school_year.leaders.add(leader)
        new.leaders.set(old.all_leaders)
        new.questions.set(old.questions.all())
        new.attachments.set(old.attachments.all())
        return new


class OrderableRegistration(SubjectRegistration):
    subject_type = SubjectType.ORDERABLE
    start_date = models.DateField(_("start date"))
    start_time = models.TimeField(_("start time"), blank=True, null=True)

    class Meta:
        app_label = "leprikon"
        verbose_name = _("orderable event registration")
        verbose_name_plural = _("orderable event registrations")

    def get_payment_status(self, d=None):
        return PaymentStatus(
            price=self.price,
            discount=self.get_discounted(d),
            explanation=",\n".join(
                discount.explanation.strip()
                for discount in self.all_discounts
                if (d is None or discount.accounted.date() <= d) and discount.explanation.strip()
            ),
            received=self.get_received(d),
            returned=self.get_returned(d),
            current_date=d or date.today(),
            due_from=self.payment_requested
            and (
                self.payment_requested.date()
                if self.subject.orderable.due_from_days is None
                else max(
                    self.start_date - timedelta(days=self.subject.orderable.due_from_days),
                    self.payment_requested.date(),
                )
            ),
            due_date=self.payment_requested
            and max(
                self.start_date - timedelta(days=self.subject.orderable.due_date_days),
                self.payment_requested.date() + timedelta(days=self.subject.min_due_date_days),
            ),
        )

    @cached_property
    def end_date(self):
        if self.start_time:
            return (datetime.combine(self.start_date, self.start_time) + self.subject.orderable.duration).date()
        else:
            return self.start_date + self.subject.orderable.duration

    @cached_property
    def end_time(self):
        if self.start_time:
            return (datetime.combine(self.start_date, self.start_time) + self.subject.orderable.duration).time()

    def event_date(self):
        return "{start}{separator}{end}".format(
            start=(
                date_format(datetime.combine(self.start_date, self.start_time), "SHORT_DATETIME_FORMAT")
                if self.start_time
                else date_format(self.start_date, "SHORT_DATE_FORMAT")
            ),
            separator=" - " if self.start_date != self.end_date or self.end_time is not None else "",
            end=(
                (time_format(self.end_time, "TIME_FORMAT") if self.end_time else "")
                if self.start_date == self.end_date
                else (
                    date_format(datetime.combine(self.end_date, self.end_time), "SHORT_DATETIME_FORMAT")
                    if self.end_time
                    else date_format(self.end_date, "SHORT_DATE_FORMAT")
                )
            ),
        )

    event_date.admin_order_field = "start_date"
    event_date.short_description = _("event date")


class OrderableDiscount(SubjectDiscount):
    registration = models.ForeignKey(
        OrderableRegistration, on_delete=models.CASCADE, related_name="discounts", verbose_name=_("registration")
    )

    class Meta:
        app_label = "leprikon"
        verbose_name = _("orderable event discount")
        verbose_name_plural = _("orderable event discounts")
        ordering = ("accounted",)


class OrderablePlugin(CMSPlugin):
    event = models.ForeignKey(Orderable, on_delete=models.CASCADE, related_name="+", verbose_name=_("event"))
    template = models.CharField(
        _("template"),
        max_length=100,
        choices=settings.LEPRIKON_ORDERABLE_TEMPLATES,
        default=settings.LEPRIKON_ORDERABLE_TEMPLATES[0][0],
        help_text=_("The template used to render plugin."),
    )

    class Meta:
        app_label = "leprikon"


class OrderableListPlugin(CMSPlugin):
    school_year = models.ForeignKey(
        SchoolYear, blank=True, null=True, on_delete=models.CASCADE, related_name="+", verbose_name=_("school year")
    )
    departments = models.ManyToManyField(
        Department,
        blank=True,
        related_name="+",
        verbose_name=_("departments"),
        help_text=_("Keep empty to skip searching by departments."),
    )
    event_types = models.ManyToManyField(
        SubjectType,
        blank=True,
        limit_choices_to={"subject_type": SubjectType.ORDERABLE},
        related_name="+",
        verbose_name=_("event types"),
        help_text=_("Keep empty to skip searching by event types."),
    )
    age_groups = models.ManyToManyField(
        AgeGroup,
        blank=True,
        related_name="+",
        verbose_name=_("age groups"),
        help_text=_("Keep empty to skip searching by age groups."),
    )
    target_groups = models.ManyToManyField(
        TargetGroup,
        blank=True,
        related_name="+",
        verbose_name=_("target groups"),
        help_text=_("Keep empty to skip searching by target groups."),
    )
    groups = models.ManyToManyField(
        SubjectGroup,
        blank=True,
        related_name="+",
        verbose_name=_("event groups"),
        help_text=_("Keep empty to skip searching by groups."),
    )
    leaders = models.ManyToManyField(
        Leader,
        verbose_name=_("leaders"),
        blank=True,
        related_name="+",
        help_text=_("Keep empty to skip searching by leaders."),
    )
    template = models.CharField(
        _("template"),
        max_length=100,
        choices=settings.LEPRIKON_ORDERABLELIST_TEMPLATES,
        default=settings.LEPRIKON_ORDERABLELIST_TEMPLATES[0][0],
        help_text=_("The template used to render plugin."),
    )

    class Meta:
        app_label = "leprikon"

    def copy_relations(self, oldinstance):
        self.departments.set(oldinstance.departments.all())
        self.event_types.set(oldinstance.event_types.all())
        self.groups.set(oldinstance.groups.all())
        self.age_groups.set(oldinstance.age_groups.all())
        self.target_groups.set(oldinstance.age_groups.all())
        self.leaders.set(oldinstance.leaders.all())

    @cached_property
    def all_departments(self):
        return list(self.departments.all())

    @cached_property
    def all_event_types(self):
        return list(self.event_types.all())

    @cached_property
    def all_groups(self):
        return list(self.groups.all())

    @cached_property
    def all_age_groups(self):
        return list(self.age_groups.all())

    @cached_property
    def all_target_groups(self):
        return list(self.target_groups.all())

    @cached_property
    def all_leaders(self):
        return list(self.leaders.all())

    Group = namedtuple("Group", ("group", "objects"))

    def render(self, context):
        school_year = (
            self.school_year or getattr(context.get("request"), "school_year") or SchoolYear.objects.get_current()
        )
        events = Orderable.objects.filter(school_year=school_year, public=True).distinct()

        if self.all_departments:
            events = events.filter(department__in=self.all_departments)
        if self.all_event_types:
            events = events.filter(subject_type__in=self.all_event_types)
        if self.all_age_groups:
            events = events.filter(age_groups__in=self.all_age_groups)
        if self.all_target_groups:
            events = events.filter(target_groups__in=self.all_target_groups)
        if self.all_leaders:
            events = events.filter(leaders__in=self.all_leaders)
        if self.all_groups:
            events = events.filter(groups__in=self.all_groups)
            groups = self.all_groups
        elif self.all_event_types:
            groups = SubjectGroup.objects.filter(subject_types__in=self.all_event_types)
        else:
            groups = SubjectGroup.objects.all()

        context.update(
            {
                "school_year": school_year,
                "events": events,
                "groups": (self.Group(group=group, objects=events.filter(groups=group)) for group in groups),
            }
        )
        return context


class FilteredOrderableListPlugin(CMSPlugin):
    school_year = models.ForeignKey(
        SchoolYear, blank=True, null=True, on_delete=models.CASCADE, related_name="+", verbose_name=_("school year")
    )
    event_types = models.ManyToManyField(
        SubjectType,
        limit_choices_to={"subject_type": SubjectType.ORDERABLE},
        related_name="+",
        verbose_name=_("event types"),
    )

    class Meta:
        app_label = "leprikon"

    def copy_relations(self, oldinstance):
        self.event_types = oldinstance.event_types.all()

    @cached_property
    def all_event_types(self):
        return list(self.event_types.all())

    def render(self, context):
        school_year = (
            self.school_year or getattr(context.get("request"), "school_year") or SchoolYear.objects.get_current()
        )

        from ..forms.subjects import SubjectFilterForm

        form = SubjectFilterForm(
            subject_type_type=SubjectType.ORDERABLE,
            subject_types=self.all_event_types,
            school_year=school_year,
            is_staff=context["request"].user.is_staff,
            data=context["request"].GET,
        )
        context.update(
            {
                "school_year": school_year,
                "form": form,
                "events": form.get_queryset(),
            }
        )
        return context



from django import template

from ..rocketchat import get_rc_id, get_rc_url, get_rc_ws_url

register = template.Library()


@register.inclusion_tag("rocketchat/chat.html", takes_context=True)
def chat(context):
    user = getattr(context.get("request"), "user")
    return {
        "rocketchat_url": get_rc_url(),
        "websocket_url": get_rc_ws_url(),
        "user": user,
        "user_id": user and user.is_authenticated and get_rc_id(user),
    }


@register.inclusion_tag("rocketchat/livechat.html", takes_context=True)
def livechat(context):
    return {
        "rocketchat_url": get_rc_url(),
        "user": getattr(context.get("request"), "user"),
    }
